# todisc.py

"""A GUI for the todisc command-line program.
"""

try:
    import Tkinter as tk
except ImportError:
    # python 3
    import tkinter as tk

# Get supporting classes from libtovid.metagui
from libtovid.metagui import *
from libtovid.guis.helpers import *


"""Since todisc has a large number of options, it helps to store each
one in a variable, named after the corresponding command-line option.

Below are defined all the todisc options that will appear in the GUI.
The actual GUI layout comes afterwards...
"""

### --------------------------------------------------------------------
### Main todisc options
### --------------------------------------------------------------------

# Note: _files_and_titles is defined in helpers.py

_heading = Label(heading_text, 'center', tovid_icon, 48, 48)

_menu_title = Text('Menu title', '-menu-title', 'My video collection',
    'Title text displayed on the main menu. Use \\n for a new line '
     'in a multi-line title.')

_group = ListToMany('-files', 'Grouped videos', '-group', None,
    'Video files to group together. Select the video in the list ' \
    'on the left, and add files to group with it by using the file ' \
    'selector on the right',
    Filename('', filetypes=video_filetypes),
    index=True,
    repeat=True)

_ntsc = Flag('NTSC', '-ntsc', True, 'NTSC, US standard')

_pal = Flag('PAL', '-pal', False, 'PAL, European standard')


### --------------------------------------------------------------------
### Showcase options
### --------------------------------------------------------------------

_non_showcase = Label(
    'Default is centered thumbnail menu links.\nOr choose '
    'from the following "Edge aligned styles".\n'
    '( see tooltips )')

_showcase = FlagOpt('Showcase', '-showcase', False,
    'Arrange menu links along the outside edges, to leave room for'
    ' an optional "showcase" image or video.  The file entry box'
    ' is for an optional image or video file to be showcased in a'
    ' large central frame',
    Filename('', action='load', desc='Select an image or video file.',
    filetypes=visual_filetypes))

_showcase_seek = Number('Showcase seek', '-showcase-seek', 2,
    'Play showcase video from the given seek time. '
    'Note: switched menus uses the value(s) from the '
    ' "Thumbnail seek(s)" option, not this one',
    0, 3600, 'secs')

_textmenu = FlagOpt("Textmenu", '-textmenu', False,
    'A menu using just text for links instead of image AND text, arranged at '
    'outside edges.  Optionally use a "showcase" image/video, or '
    'static/animated background.  Use the "columns" option to set the number '
    'of titles in column 1 (column 2 will use the remainder).  Note that '
    'column 2 titles are aligned right. '
    'See also "Quick menu" and "Switched menus".',
    Number('Columns', '', 13, '', 0, 13))

_quick_menu = Flag('Quick menu (may need a menu video)',
    '-quick-menu', False, 'Note: you need a recent ffmpeg with the "movie" '
    'filter enabled. Ten times faster than normal showcase animation.  A '
    'showcase or background video is required unless doing switched menus.  '
    'Menu links are text only.  Not compatible with wave or rotate options.')

_switched_menus = Flag('Switched menus (try with "Quick menu" !)',
    '-switched-menus', False,
    'This makes a showcase style menu with text menu '
    'links.  The showcased VIDEO or IMAGE will be of each '
    'video "title", and will change as you press the '
    'up/down keys on your remote.  Do not select a '
    'showcase file for this option.  Use with '
    '"Quick menu" option for a huge speed up !')

_showcase_framestyle = Choice('', '-showcase-framestyle', 'default',
    'This option is only for menu styles with a "showcase" image.  '
    'The "none" option will use the default frame method, using imagemagick.  '
    'The "glass" option will use mplayer to make frames, giving an animated '
    'effect.  The glass style can be much faster - especially if used without '
    '-rotate and -wave options',
    'default|glass')

_showcase_framesize = Number('Frame size', '-showcase-framesize', 4,
    'Width of the showcase image/video frame in pixels',
    0, 20, 'pixels')

_showcase_frame_color = Color('Frame color', '-showcase-frame-color',
    '#6E6E6E', 'Color of the showcase frame. ')

_thumb_framesize = Number('Frame size', '-thumb-frame-size', 4,
    'Width of the thumbnail link frame in pixels',
    0, 20, 'pixels')

_thumb_frame_color = Color('Frame color', '-thumb-frame-color',
    '#6E6E6E', 'Color of the thumbnail link frame. ')

_showcase_shape = Choice('Showcase shape', '-showcase-shape', 'none',
    'Apply a shaped transparency mask to showcase videos or images. '
    'No frame will be used for this option.  '
    'Leave at "none" to not use a feathered shape.',
    thumb_masks, 'dropdown')

_showcase_geo = Text('Image position', '-showcase-geo', '',
    'Enter the position of the top left corner of the showcase image: '
    'e.g. "200x80".  This value is applied to the video *before* is is scaled.')

_showcase_titles_align = Choice('Title alignment',
    '-showcase-titles-align', 'none',
    'This option is only for showcase style menus with video thumbnails and '
    'text.  Default is to center the text above the thumbnails.  This option '
    'will align the titles either to the left (west), center, or right '
    '(east).  Leave at "none" to let todisc sort it out for you.',
    'none|west|center|east')

# Menu settings
_bg_color = Color('', '-bg-color', '#101010',
    'Color of the menu background. Default (#101010) is NTSC color-safe black. '
    'Note: the color turns out MUCH lighter than the one you choose, '
    'so pick a VERY dark version of the color you want.')

_background = Filename('File', '-background', '',
    'Image or video displayed in the background of the main menu',
    'load', 'Select an image or video file',
    filetypes=visual_filetypes)

_bgaudio = Filename('File', '-bgaudio', '',
    'Audio file to play while the main menu plays.  '
    'Static menus use default audio length of 20 seconds.  '
    'Change with "Menu length" on "Menu" tab.  '
    'Use almost any filetype containing audio.',
    'load', 'Select a file containing audio', filetypes=av_filetypes)

_bgvideo_seek = Number('Seek', '-bgvideo-seek', 2,
    'Play background video from the given seek time (seconds)',
    0, 3600, 'secs')

_bgaudio_seek = Number('Seek', '-bgaudio-seek', 2,
    'Play background audio from the given seek time.',
    0, 3600, 'secs')

_menu_audio_fade = Number('Fade in', '-menu-audio-fade', 1,
    'Number  of  sec to fade given menu audio in and out '
    '(default: 1.0 seconds). Use a fade of "0" for no fade.',
    0, 10, 'secs')

_menu_fade = FlagOpt('Menu fade (in/out)', '-menu-fade', False,
    'Fade the menu in and out. The background will fade in first, then '
    'the title (and mist if called for), then the menu thumbs and/or titles.  '
    'The fadeout is in reverse order.  The optional numerical argument '
    'is the length of time the background will play before the menu '
    'begins to fade in.  To disable the fadeout portion, set the '
    '"Pause indefinitely" flag on the "Playback" tab.',
    Number('Start', '', 1, '', 0, 60, 'secs'))

_transition_to_menu = Flag('Transition to menu', '-transition-to-menu', False,
    'A convenience option for animated backgrounds using a menu fadein: the '
    'background will become static at the exact point the thumbs finish '
    'fading in.  This menu does not loop '
    'unless you enable "Menu looping" on the "Playback" tab).')

_intro = Filename('Intro video', '-intro', '',
    'Video to play before showing the main menu.  At present this must '
    'be a DVD compatible video at the correct resolution etc.  Only 4:3 '
    'aspect is supported.',
    'load', 'Select a video file',
    filetypes=[('DVD video', dvd_video_files)])

_menu_length = Number('Menu length', '-menu-length', 20,
    'Duration of menu. The length of the menu will also set '
    'the length of background audio for a static menu',
    0, 120, 'secs')


# Static/animated main menu
_static = Flag('Static menu', '-static', False,
    'Create still-image menus; takes less time. For duration of background '
    'audio for static menus, use "Menu length" on this tab.  See "Main menu" '
    'tab for other menu styles and options.')

_animated = Flag('Animated', '', True,
    'Create animated menus.  See "Main menu" tab for other menu styles and '
    'options.')

# Static/animated submenus
_submenus = Flag('Static submenus', '-submenus', False,
    'Create a submenu for each video title.  Submenu links lead to '
    ' chapter points.  See "Submenus" tab for more submenu options.')

_ani_submenus = Flag('Animated', '-ani-submenus', False,
    'Create an animated submenu for each video.  Submenu links lead to '
    'chapter points.  See "Submenus" tab for more submenu options.')


_submenu_length = Number('Submenu length', '-submenu-length', 14,
    'The length of the submenu. If doing static submenus and using audio '
    'for the submenu, this will be the length of the submenu audio',
    0, 80, 'secs')

_submenu_audio_fade = Number('Audio fade', '-submenu-audio-fade', 1,
    'Number of seconds to fade given submenu audio in and out.',
    0, 10, 'secs')

_submenu_audio = List('Audio', '-submenu-audio', None,
    'File(s) that will play as background audio for each submenu.  '
    'Use a single file for all submenus, or 1 file per submenu.  '
    'Any file containing audio (that ffmpeg can handle) can be used.',
    Filename('', filetypes=av_filetypes))

_submenu_background = List('Image(s)', '-submenu-background', None,
    'Background image(s) for the submenus(s). Single value or list',
    Filename('', filetypes=image_filetypes))

_submenu_bg_color = Color('Background color', '-submenu-bg-color', '#101010',
    'Background color to use for submenu(s). Default (#101010) is '
    'NTSC-safe black.')

_submenu_titles = ListToOne('-files', 'Submenu titles', '-submenu-titles', None,
    'Submenu titles for each video.  '
    'Use \\n for a new line in a multi-line title.',
    Text(),
    filter=strip_all)

_chapter_titles = ListToMany('-files', 'Chapter titles', '-chapter-titles', None,
    'Chapter titles for each video.  Use \\n for a new line in '
    'a multi-line title.  Number of titles given must equal the '
    'number of chapters given for that video.  HINT: If you '
    'click on a video title in the list to the left, then click '
    '"Add" repeatedly until you reach the desired number of '
    'chapters, you can then edit the titles from the keyboard using '
    'the Enter key to cycle through them. ',
    Text(),
    side='left',
    repeat=True)

_title_gap = Number('Space between titles', '-title-gap', 10,
    'Leave this much vertical gap between titles.  '
    'Default is 10 for line buttons, 15 for text-rect buttons.  '
    'This value is applied before the menu is scaled.',
    0, 400, 'pixels')

_text_start = Number('Start titles at', '-text-start', 50,
    'Titles will start at this pixel in the vertical axis.  '
    'This value is applied before the menu is scaled.',
    0, 460, 'pixels')

### --------------------------------------------------------------------
### Runtime behavior
### --------------------------------------------------------------------

_jobs = Number('Number of jobs to run simultaneously', '-jobs', 0,
    'Leave this value at 0 if you wish the default, which is to run '
    'as many jobs as processors found. Use this option if you wish '
    'to limit or increase this number.', 0, 32)
_keep_files = Flag('Keep intermediate files on exit', '-keep-files', False)
_no_ask = Flag('No prompts for questions', '-no-ask', False)
_no_warn = Flag('No pause at warnings', '-no-warn', False)
_grid = Flag('Grid preview', '-grid', False,
    'Show a second preview image with a grid and numbers that will help in '
    'finding coordinates for options that might use them')

_no_menu = Flag('No menu', '-no-menu', False,
    'Create a DVD with no menu. This creates a DVD with no menu '
    'which will jump to the first video, and play all videos provided '
    'in sequence.  Chapters provided with the -chapters option will be '
    'the chapter interval in minutes, or alternatively you can provide '
    'a single chapter list in HH:MM:SS format.')

_slides = List('Images for slideshow one', '-slides', None,
    "Image files for the slideshow",
    Filename('', filetypes=image_filetypes))

_submenu_slide_total = SpacedText('Number of slides shown on submenu',
    '-submenu-slide-total', '',
    'Use this many slides for making the slideshow submenus. '
    'The default is to use all the slides given.  '
    'For multiple slideshows or slideshow(s) mixed with video(s) only.  '
    'Select a submenu option on the "Basic" tab to use this option. '
    'Use a single value for all or list one value per submenu.')

_slide_transition = Choice('Transition type', '-slide-transition', 'crossfade',
    'The type of fade transition between slides in an '
    'animated slideshow menu.  Be sure  the  menu  length '
    'is long enough to support the 1 second transitions '
    'between the slides.  The length is determined by  '
    '1)  the length  of  the -bgaudio AUDIO  2) the '
    'length given with -menu-length DURATION.  '
    '* See "Basic" tab for menu length option *.',
    'crossfade|fade')

_slide_border = FlagOpt('Slide border', '-slide-border', False,
    'Use a border around animated slideshow slides (default: 100 pixels).',
    Number('width', '', 100, '', 0, 200, 'pixels'))

_slide_frame = FlagOpt('Slide frame', '-slide-frame', False,
    'Use a frame around the animated slideshow slides (default: 12).',
    Number('width', '', 12, '', 0, 20, 'pixels'))

_no_confirm_backup = Flag('No reminder prompt to  backup your slideshow files',
    '-no-confirm-backup', False, 'Todisc prompts you for a "yes" '
    ' to show you have taken backup precations.  This option '
    'disables that prompt.')

### --------------------------------------------------------------------
### Slideshows
### --------------------------------------------------------------------

_background_slideshow = Flag("Background", '-background-slideshow', False,
    'This option is only for multiple slideshows, '
    '(or mixed video and slideshow menus). It uses the '
    'animated slideshow as the background video')

_showcase_slideshow = Flag("Showcase", '-showcase-slideshow', False,
    'This option is only for multiple slideshows, '
    '(or mixed video and slideshow menus).  It uses the '
    'animated slideshow as the showcase video')

_slideshow_menu_thumbs = List('Menu slide thumbs', '-slideshow-menu-thumbs', None,
    'Slides for the menu thumbnail.  For multiple slideshows or '
    'mixed slideshows/videos only.',
    Filename('', filetypes=image_filetypes))


_use_dvd_slideshow = FlagOpt('Use dvd-slideshow program', '-use-dvd-slideshow', False,
    'Use the "dvd-slideshow" program to make the animated '
    'slideshow. You need of course to have installed this program '
    'to use this. You only need this for special effects like the '
    '"Ken Burns effect" otherwise todisc can do the job well '
    'and fast on its own. The optional file is the dvd-slideshow '
    'file you would need for non-todisc options, otherwise '
    'todisc will make the file for you. Using this option will '
    'void many todisc slideshow options.',
    Filename('', action='load', help='Select the dvd-author configuration file.'))

_slides_to_bin = List('Slides to "bin"', '-slides-to-bin', None,
    '"Binning" resizes the image to 640x480 using a "box" '
    'filter, which reduces "signal to noise ratio"'
    'Use in conjuction with the next option for burring, '
    'to achieve stronger denoising',
    Filename('', filetypes=image_filetypes))

_slides_to_blur = List('Slides to blur', '-slides-to-blur', None,
    'Blurring the image can reduce noise such as "ringing" '
    'effects.  Use values between 0x0.1 and 0x0.9 for best '
    'results.  Enter a single value or 1 value for each slide '
    'to be blurred.',
    Filename('', filetypes=image_filetypes))

_slide_blur = SpacedText('Slide blur amount', '-slide-blur', '',
    'How much to blur each slide given with "Slides to blur".  '
    'Use a single value, or a list with one value per slide given'
    'The format is {radius}x{sigma} and the default is 0x0.2.  Using '
    'values between 0x0.1 and 0x0.9 is probably the best range.')


### --------------------------------------------------------------------
### Thumbnail images
### --------------------------------------------------------------------

_thumb_shape = Choice('Thumb shape', '-thumb-shape', 'none',
    'Apply a shaped transparency mask to thumbnail videos or images.  These '
    '"feathered" shapes look best against a plain background or used with '
    '-thumb-mist [COLOR].  To use a "mist" background behind each thumb, '
    'see "Thumb mist" section.  No frame will be used for this option.  '
    'See "man tovid" (-thumb-shape) for info on how to add your own masks.'
    'Leave at "none" to not use a feathered shape.',
    thumb_masks, 'dropdown')

_thumb_columns =  Choice('Thumb columns', '-thumb-columns', 'none',
    'Use a montage tile of 3x1 or 4x1 instead of the usual 2x2 for 3 videos.  '
    'If you want 1 row with 3 thumbs choose "3", or for 1 row with 4 thumbs '
    'choose "4".',
    'none|3|4', 'dropdown')

_opacity = Number('Opacity', '-opacity', 100,
    'Opacity  of thumbnail videos as a percentage. '
    'Less than 100(%) is semi-transparent. '
    'Not recommended with dark backgrounds.',
    1, 100, '%')

_blur = Number('Blur', '-thumb-blur', 1.0,
    'The amount of feather blur to apply to the thumb shape.  '
    'Default is 1.0 which will more or less keep the shape, creating "soft" '
    'edges.  Use float or integer values between 0.1 and 2.0',
    0.1, 2.0, '', 'popup', 0.1)

_showcase_blur = Text('Blur', '-showcase-blur', "",
    'The amount of feather blur to apply to the showcase thumb shape.  '
    'Default is 1.0 which will more or less keep the shape, creating "soft" '
    'edges.  Use float or integer values between 0.1 and 2.0')

_3dthumbs = Flag('3D thumbs', '-3d-thumbs', False,
    'This will give an illusion of 3D to the thumbnails: '
    'dynamic lighting on rounded thumbs, and a  raised '
    ' effect on rectangular thumbs')

_rotate_thumbs = SpacedText('Rotate Thumbs (list)', '-rotate-thumbs', '',
    'Rotate  thumbs  the  given amount in degrees - can be positive or '
    'negative.  There must be one value for each file given with files.  If '
    'the values are not the same distance from zero, the thumbs will be of '
    'different sizes as images are necessarily resized *after* rotating.  '
    'Note: this will not change a portrait image into a landscape image!')

_wave = Flag('Wave effect', '-wave', False,
    'Wave effect for showcase image|video.  Alters thumbs along a sine '
    'wave.  This will pass a wave arg of -20x556, producing a gentle '
    'wave with a small amount of distortion.  To use other values you '
    'will need to use "-wave VALUE" in the "todiscopts" box on the "Behavior" '
    'tab.  See man todisc for details.')

_3dshowcase = Flag('3D thumb', '-3d-showcase', False,
    'This will give an illusion of 3D to the thumbnails: '
    'dynamic lighting on rounded thumbs, and a  raised '
    ' effect on rectangular thumbs')

_rotate = Number('Rotate thumb', '-rotate', 0,
    'Rotate the showcase image|video clockwise by this number of degrees.'
    'Note: this will not change a portait image into a landscape image!',
    -30, 30, 'degrees')

_thumb_mist = FlagOpt('Use thumb mist', '-thumb-mist', False,
    'Use a mist behind shaped thumbnails for contrast with the background.  '
    'The Color option is the color of the mist.  With some font and mist '
    'color combos you may need to use a large bold font for readability.',
    Color('', '', '#ffffff', ''))

_tile3x1 = Flag('1 row of 3 thumbs', '-tile3x1', False,
    'Use a montage tile of 3x1 instead of the usual 2x2 '
    '(3 videos only).  Not a showcase option.')

_tile4x1 = Flag('1 row of 4 thumbs', '-tile4x1', False,
    'Use a montage tile of 4x1 instead of the usual 2x2 '
    '(4 videos only).  Not a showcase option.')

_align = Choice('Align', '-align', 'north',
    'Controls positioning of the thumbnails (if any) and video titles.  '
    'With some arrangements this will have limited effects however.'
    'For example only north|south|center can be used with the default montage '
    'of thumbs arrangement, or with showcase with thumb arrangement. '
    'It will be most effective with single column showcase and textmenu styles',
    'north|south|east|west|center', 'dropdown')

_seek = SpacedText('Thumbnail seek(s)', '-seek', '',
    'Play thumbnail videos from the given seek time (seconds).  '
    'A single value or space separated list, 1 value per video.  '
    'Also used for seeking in switched menus.')

_user_thumbs = List('Image(s)', '-user-thumbs', None,
    'Images for thumbnails.  This option requires one image for each title.',
    Filename('', filetypes=image_filetypes))

### --------------------------------------------------------------------
### Fonts and font colors
### --------------------------------------------------------------------

# Menu title font
_menu_font = Font('', '-title-font', 'Helvetica',
    'The font to use for the menu title')

_menu_fontsize = Number('Size', '-title-fontsize', 24,
    'The font size to use for the menu title',
    0, 80, 'pixels', toggles=True)

_title_color = Color('', '-title-color', '#eaeaea',
    'The font color to use for the menu title')

_title_opacity = Number('Opacity', '-title-opacity', 100,
    'The opacity of the menu title.',
    0, 100, '%')

_title_stroke = Color('Stroke', '-title-stroke', None,
    'Outline color for the main menu font.')

menu_title_font = HPanel('Menu title font',
    _menu_font,
    _menu_fontsize,
    _title_color,
    _title_opacity,
    _title_stroke,
)

# Video title font
_titles_font = Font('', '-titles-font', 'Helvetica',
    'The font to use for the video titles')
# unused FIXME
_titles_font_deco = SpacedText('Custom font decoration', '', '',
         'Space-separated list of custom options to imagemagick.'),

_titles_fontsize = Number('Size', '-titles-fontsize', 12,
    'The font size to use for the video titles.  '
    'Default size varies depending on options chosen.',
    0, 80, 'pixels', toggles=True)

_titles_color = Color('', '-titles-color', None,
    'The font color to use for the video titles')

_titles_opacity = Number('Opacity', '-titles-opacity', 100,
    'The opacity of the video titles.',
    0, 100, '%')

_titles_stroke = Color('Stroke', '-titles-stroke', None,
    'The color to use for the video titles font outline (stroke)')

video_title_font = HPanel('Video title font',
    _titles_font,
    _titles_fontsize,
    _titles_color,
    _titles_opacity,
    _titles_stroke,
)

# Submenu font
_submenu_font = Font('', '-submenu-font', 'Helvetica',
    'The font to use for the Submenu menu titles')

_submenu_fontsize = Number('Size', '-submenu-fontsize', 24,
    'The font size to use for the submenu title',
     0, 80, toggles=True)

_submenu_title_color = Color('', '-submenu-title-color', '#eaeaea',
    'The font color to use for submenu title(s)')

_submenu_title_opacity = Number('Opacity',
    '-submenu-title-opacity', 100,
    'The opacity of the submenu title.',
    0, 100, '%')

_submenu_stroke = Color('Stroke', '-submenu-stroke', None,
    'The color for the submenu font outline (stroke).')

submenu_title_font = HPanel('Submenu title font',
    _submenu_font,
    _submenu_fontsize,
    _submenu_title_color,
    _submenu_title_opacity,
    _submenu_stroke,
)

# Submenu chapter font
_chapter_font = Font('', '-chapter-font', 'Helvetica',
    'The font to use for the chapter titles')

_chapter_fontsize = Number('Size', '-chapter-fontsize', 12,
    'The font size to use for the chapter titles',
    0, 80, 'pixels', toggles=True)

_chapter_color = Color('', '-chapter-color', None,
    'The color for the chapters font.')

_chapter_title_opacity = Number('Opacity',
    '-chapter-title-opacity', 100,
    'The opacity of the chapter titles.',
    0, 100, '%')

_chapter_stroke = Color('Stroke', '-chapter-stroke', None,
    'The color for the chapters font outline (stroke).')

submenu_chapter_font = HPanel('Submenu chapter font',
    _chapter_font,
    _chapter_fontsize,
    _chapter_color,
    _chapter_title_opacity,
    _chapter_stroke,
)

# Other stuff
_menu_title_geo = Choice('Title position', '-menu-title-geo', 'none',
    'The position of the menu title',
    'north|south|west|east|center|none', 'dropdown')

_menu_title_offset = Text('Title offset', '-menu-title-offset', '+0+50',
    'Menu title position as an offset (in pixels) from '
    'the N/S/E/W "Title position".  Use +X+Y syntax.  This '
    'value is applied to the video *before* is is scaled.')

_text_mist = Flag('Text mist', '-text-mist', False,
    'Use "mist" behind the menu title (helps with contrast).',
    enables=['-text-mist-color', '-text-mist-opacity'])

_text_mist_color = Color('', '-text-mist-color', '#ffffff',
    'Color of the mist behind the menu title.')

_text_mist_opacity = Number('Opacity', '-text-mist-opacity', 60,
    'The opacity of the mist behind the menu title.',
    1, 100, '%')




### --------------------------------------------------------------------
### Menu buttons (spumux)
### --------------------------------------------------------------------

_button_style = Choice('Style', '-button-style', 'none',
    'Style or shape of the buttons seen when playing the DVD '
    'menu.  "rect": rectangle around the thumbs, "text": uses '
    'the title text, "text-rect": rectangle around the title text, '
    '"line": line underneath title.  The "line" style is showcase only'
    ', and the "text" buttons can not be used with "showcase" menus.'
    'Leave at "none" to let todisc pick the most appropriate style. ',
    'none|rect|line|text|text-rect', 'dropdown')

_highlight_color = Color('Highlight', '-highlight-color', '#266cae',
    'Color for the menu buttons the dvd remote uses to navigate.')

_select_color = Color('Selection', '-select-color', '#de7f7c',
    'Color for the menu buttons the dvd remote uses to select.')

_outlinewidth = Number('Outlinewidth', '-outlinewidth', 14,
    'For spumux outlinewidth variable.  This option helps if spumux '
    'fails because of a large gap between button words or characters.',
    0, 20, 'pixels')

_playall = Flag('"Play all" button', '-playall', False,
    'Create a "Play All" button that jumps to the 1st title and plays '
    'all the videos in succession before returning to the main menu.')

_chapters = Chapters('-files', 'Chapters', '-chapters', None,
    'Number of chapters or HH:MM:SS string for each video. '
    'If only one value is given, use that for all videos. '
    'For grouped videos, use a "+" separator for joining '
    'the chapter string of each grouped video. '
    'An example chapter string using HH:MM:SS format: '
    '00:00:00,00:05:00,00:10:00.  An example of 2 grouped videos: '
    '00:00:00,00:05:00+00:05:00,00:10:00.  '
    'When using HH:MM:SS format the 1st chapter MUST be 00:00:00.  '
    'If using -no-menu and passing just integer(s), then the value '
    'represents the chapter INTERVAL not the number of chapters',
    text='set with mplayer',
    side='top',
    filter=strip_all)

_loop = FlagOpt('Menu looping (pause)', '-loop', False,
    'Pause before looping playback of the main menu.  Set the number spinbox '
    'to the pause desired.  Default pause varies depending on other options '
    'chosen.  Use -1 to disable looping',
    Number('', '', '', '', -1, 30, 'secs'))

_chain_videos = FlagOpt('Chain videos', '-chain-videos', False,
    'This option will "chain" videos so they play '
    'sequentially until the last, which will return to '
    'the menu.  You can also specify individual videos to '
    'behave like this by using the text box at the right. '
    'You can use single integers or indicate a range with a '
    '"-" separator. Example of a range: "1-2 4-6" (no quotes). '
    'This will play video 1 through video 3 before returning '
    'to the menu, and the same for video 4 through video 6.',
    Text(''))

_videos_are_chapters = Flag('Each video is a chapter',
    '-videos-are-chapters', False,
    'A button will be made on the main menu for each video which You '
    'can use as a chapter button .  Selecting any video will play '
    'them all in order starting with the selected one.' )


### --------------------------------------------------------------------
### Authoring / burning
### --------------------------------------------------------------------

# Burning options
_burn = Flag('Burn project', '-burn', False, 'Burn project on completion')
_eject = Flag('Eject', '-eject', False, 'Eject DVD when burn complete')

_speed = Number('Speed', '-speed', 8,
    'Speed for burning',
    0, 30)

_device = Text('Device', '-device', '/dev/dvdrw',
    'Type your burning device (default: /dev/dvdrw)')

# Authoring
_widescreen = Choice('Widescreen', '-widescreen', 'none',
    'This will output a <video widescreen=nopanscan /> tag '
    '(for example) for the dvdauthor xml file.  It affects all '
    'videos in the titleset.  Use in conjunction with "Video aspect ratio" '
    'if your dvd player is cropping your videos.  '
    'Leave this at "none" to not output a widescreen tag',
    'none|nopanscan|noletterbox')

_aspect = Choice('Video aspect ratio', '-aspect', 'none',
    'This will output a <video aspect WIDTH:HEIGHT /> tag for the '
    'dvdauthor xml file.  It will affect all videos in the titleset.  '
    'This also has the effect of forcing a thumb aspect ratio for the menu, '
    'which is otherwise determined automatically. '
    'Leave this at "none" to let todisc and dvdauthor figure it out for you.',
    '4:3|16:9|none')

_audio_lang = SpacedText('Audio', '-audio-lang', '',
    'Single value or list of language codes for audio. '
    'Identify the audio tracks on the DVD.  These language codes are used for '
    'each video in the titleset.  When you use the audio button on your DVD '
    'remote the language name is displayed. Example: -audio-lang en fr ')

_audio_channel = SpacedText('Audio channel', '-audio-channel', '',
    'Space-separated list of audio channel numbers, for example '
    '"Video1_track Video2_track Video3_track", where "VideoN_track" is the '
    'track number to use in a multi-track (multi-language) mpeg: usually '
    'something like -audio-channel "1 0 1".  The '
    '1st track is 0, 2nd is 1 . . .  etc. If the tracks are 0.English '
    '1.French, then the above would make French the audio language on Video1 '
    'and Video3, and English the audio language on Video2.  You can check the '
    'mpeg with "mplayer -v . . ."')

_subtitles = SpacedText('Subtitles', '-subtitle-lang', '',
    'This allows selectable subtitles in the DVD, assuming you have optional '
    'subtitles muxed into your videos. Enter a single 2-character language code, '
    'or several separated by spaces.')

_quick_nav = Flag('Quick-nav', '-quick-nav', False,
    'This option will allow navigation of a menu with more than one titleset by '
    'using the  left and  right  arrow  keys of your DVD remote.  When you press '
    'this key the highlight will go the next or previous title.  If you are at '
    'the end of a titleset the right key will go  to the next titleset.  If you '
    'are at the beginning of a titleset, the left key will go to the previous '
    'titleset.  If no next or previous titleset it will cycle to the end or '
    'beginning of the titlesets. ')

### ---------------------------------------------------------------------------
### Main GUI layout and panel construction
### ---------------------------------------------------------------------------
"""The todisc GUI is laid out using panels and tabs for logical grouping.
"""

main =  VPanel('Basic',
    _heading,_files_and_titles,
    VPanel('',
        HPanel('',
            VPanel('Menu options',
                FlagGroup('', 'exclusive',
                    _static, _animated, _no_menu, side='left'),
                FlagGroup('', 'exclusive',
                    _submenus, _ani_submenus, side='left')
            ),
            FlagGroup('TV System', 'exclusive', _ntsc, _pal),
            VPanel('Chapters', _chapters),
            VPanel('Burning',
                HPanel('',
                _burn,
                _eject),
                HPanel('',
                    _speed,
                    _device,
                ),
            ),
        ),
        HPanel('',
            HPanel('Video or image background', _background),
            HPanel('Audio background', _bgaudio),
        ),
        HPanel('', _menu_title, _menu_length),
        _out,
    ),
)

main_menu = Tabs('Main menu',
    HPanel('Basic settings',
        VPanel('',
            HPanel('Default style',  _non_showcase),
            VPanel('Edged aligned styles',
                _showcase,
                _textmenu,
                _quick_menu,
                _switched_menus,
            ),
        ),
        VPanel('',
            VPanel('Special menu style options',
                _menu_fade,
                _transition_to_menu,
                _intro,
            ),
            VPanel('Backgrounds',
                VPanel('Image or video options',
                    HPanel('', _bgvideo_seek,
                    _bg_color)),
                VPanel('Audio options',
                    HPanel('',  _bgaudio_seek, _menu_audio_fade)),
                HPanel('Menu alignment', _align),
            ),
        ),
    ),

    VPanel('Menu title',
        menu_title_font,
        HPanel('Layout',
            _menu_title_geo,
            _menu_title_offset,
        ),
        HPanel('Mist',
            _text_mist,
            _text_mist_color,
            _text_mist_opacity,
        ),
    ),

    VPanel('Video titles',
        video_title_font,
        HPanel('Layout',
            VPanel('For textmenu style titles',
            _text_start,
            _title_gap),
            VPanel('For showcase style menu with thumbs',
            _showcase_titles_align)),
        HPanel('DVD buttons',
            Label('Configure the style and color for the menu link '
            '"buttons" on the "Playback" tab')),
    ),
)

submenus = Tabs('Submenus',
    VPanel('Settings',
        Label('Enable a submenu option on the "Basic" tab to use submenus',
        'center'),
        _submenu_length,
        _submenu_audio_fade,
        _submenu_bg_color,
        HPanel('Backgrounds',
            _submenu_audio,
            _submenu_background,
        ),
    ),

    VPanel('Submenu titles',
        submenu_title_font,
        _submenu_titles,
    ),

    VPanel('Chapter titles',
        submenu_chapter_font,
        _chapter_titles,
        HPanel('Chapters',
            Label('You will find chapter settings on the'
             '"Playback" tab under the "Chapters" tab', 'center')),
    ),
)


tab_list = []
slideshow_panel = Tabs('Slideshow',
    VPanel('General options',
        Label('Make a single slideshow on this tab.  Add slideshows using '
        'numbered tabs.  On the\n"Basic" tab: mix with videos if desired, '
        'and please set "Output name" before executing.', 'center'),
        Text('Title', '-titles', '', 'Title for this slideshow.  '
        'Use a title ONLY for multiple slideshows '
        'or slideshow(s) mixed with videos'),
        _slides,
        FlagGroup("", 'exclusive',
            Flag('Animated with transitions', '', True,
                'This is the default if you are not doing a static '
                'menu.  The slides transition one to the next, using '
                'the type of transition selected in "Transition type"'),
            Flag('Static with "Polaroid stack" menu', '-static', False,
                'This puts the  slides  onto  the  background in '
                '"random" locations with random rotations, making it '
                'look like a scattered stack of photos. This is the '
                'same as setting -static on the Main tab.'), side='left'
        ),
        VPanel('Animated slideshow options',
            HPanel('',
                _slide_transition,
                HPanel('', _slide_border, _slide_frame),
            ),
        ),
        HPanel('',
            Number('Number of slides shown on menu', '-menu-slide-total', 0,
                'Use this many slides for making the slideshow menu. '
                'Leave at 0 for default: use all slides given.',
                0, 100),
            _submenu_slide_total,
        ),
    ),
    VPanel('Advanced options',
        FlagGroup('Use slideshow as background or showcase video ',
            'exclusive',
            _background_slideshow, _showcase_slideshow
        ),
        _slideshow_menu_thumbs,
        _use_dvd_slideshow,
    ),
    VPanel('Denoising options',
        HPanel('', _slides_to_bin, _slides_to_blur),
        _slide_blur),
)

tab_list.append(slideshow_panel)

for num in range(2,  13):
    title_str = 'Images for slideshow ' + str(num)
    next_panel = VPanel(str(num),
        Text('Title', '-titles', '',
            'Title for this slideshow. Leave empty if doing a single slideshow.'),
        List(title_str, '-slides', None,
            'Image files for the slideshow.',
            Filename('')),
        )
    tab_list.append(next_panel)

slideshows = Tabs('Slideshows', *tab_list)


thumbnails = Tabs("Thumbnails",
    VPanel('Menu link thumbnails',
    VPanel('',
    HPanel('',
        VPanel('',
            HPanel('',
            VPanel('Aspect ratio', Label('Automatic: force video ratio on "Playback" tab')),
            VPanel('Seeking', _seek),
            ),
            HPanel('',
            VPanel("Effects",
                HPanel('',_opacity, Label('(Also affects showcase thumb)')),
                HPanel('', _blur, _3dthumbs),
                _rotate_thumbs,
                _thumb_mist),
            VPanel('Arrangement',
                _thumb_shape,
                _thumb_framesize,
                _thumb_frame_color,
                _thumb_columns,
            ),
                ),
        ),
    ),
    VPanel('User supplied thumbs',
    VPanel('',
    Label("Use this only if you want to substitute your own thumbs for default thumbnails.", 'center'),
    _user_thumbs,
    ),
    ),
    ),
    ),
        VPanel("Showcase thumbnail",
            HPanel('Seeking', _showcase_seek),
            HPanel('',
            VPanel('Effects',
                _wave,
                HPanel('', _showcase_blur, _3dshowcase),
                _rotate,
                HPanel('', Label('Frame style'),_showcase_framestyle)),
            VPanel('Arrangement',
                _showcase_shape,
                _showcase_framesize,
                _showcase_frame_color,
                _showcase_geo),
        ),
        ),
)

from libtovid.guis import tovid

behavior = VPanel("Behavior",
    VPanel('',
        VPanel('Execution',_jobs),
        VPanel('Interaction',_keep_files, _no_ask, _no_warn, _no_confirm_backup),
        VPanel('Preview',_grid)),
    SpacedText('Custom todisc options', '', '',
         'Space-separated list of custom options to pass to todisc.'),
)

playback = Tabs("Playback",
    VPanel('Basic settings',
        HPanel('', _aspect,  _widescreen),
        HPanel('Menu Pause', _loop),
        HPanel('Language(s)', _audio_lang, _audio_channel, _subtitles),
        HPanel('Navigation',
            _playall,
            _quick_nav,
            _videos_are_chapters,
            _chain_videos,
        ),
        HPanel('DVD buttons',
            _button_style,
            _highlight_color,
            _select_color,
            _outlinewidth,
        ),
    ),
    VPanel('Grouped Videos', _group),
)

encoding = VPanel('Encoding',
    Label("\nVideo re-encoding options - you may leave these at defaults.", 'center'),
    Tabs('',
        tovid.BASIC_OPTS,
        tovid.VIDEO,
        tovid.AUDIO,
        tovid.BEHAVIOR,
    ),
    Flag('Encode only', '-encode-only', False,
    'Use this GUI for encoding only.  On the Main tab, load files and '
    'select an out filename'),
    SpacedText('Custom makempg ("tovid mpg") options', '', '',
         'Space-separated list of custom options to pass to makempg.'),
    
)

### --------------------------------------------------------------------

def run(args=None, position='', project=''):
    from libtovid.guis.helpers import get_loadable_opts, load_script
    import os
    # if the first arg is a text file, treat it as a script and try to load
    # the options from it as a list into args
    # load_script from helpers overrides load_script from gui.py as the latter
    # suffers from the load_args bug mentioned in that file.
    if args:
        script = args[0]
        if not script.startswith('-'):
            try:
                from commands import getoutput 
            except ImportError:
            # python 3
                from subprocess import getoutput 
            if os.path.exists(script) \
              and 'text' in getoutput('file %s' %script):
                script = args.pop(0)
                args.extend(load_script(script))
    # now check if we have unloadable options, this first call does only that
    # this will return 2 lists, [0] is loadable and [1] is unloadable
    # probably don't want to do this anyway, as filter_args will just return
    # the args it is sent if none are unloadable
    a = get_loadable_opts(args)
    if a[1]:
        from libtovid.guis.helpers import filter_args
        r = tk.Tk()
        args = filter_args(r, args)
    # workaround for netbooks (could use xrandr perhaps, but this works fine)
    gui_width, gui_height = 800, 660
    root = tk.Tk()
    root.withdraw()
    root.update_idletasks()
    screen_height = root.winfo_screenheight()
    # insure screen is not 800x600 or somesuch which doesn't work for gui anyway
    # need float like 1.3333333 that we multiply * 100 and convert back to int
    screen_width = float(root.winfo_screenwidth())
    if int(screen_height) < 660 and not int(screen_width / screen_height * 100) == 133:
        gui_height = 600
    root.destroy()

    # finally, run it
    app = Application('todisc',
        main, main_menu, submenus, thumbnails, slideshows, playback, behavior, encoding)
    gui = GUI("tovid gui", gui_width, gui_height, app, icon=tovid_icon, position=position)
    gui.run(args, script=project)

if __name__ == '__main__':

    try:
        run()

    except:
        import traceback
        traceback.print_exc(10)




"""This module is for parsing manual pages and extracting necessary information
for creating a metaGUI.

Needed capabilities:

* Retrieve text of a given manpage
* Read manpage for a given program and create a list of command-line options,
  with expected parameters and accompanying paragraph(s) of documentation.


"""

__all__ = [
    'get',
    'parse',
]

import re
import os
import textwrap
from libtovid.metagui.control import *

class Matcher:
    def search(self, pattern, text):
        self.value = re.search(pattern, text)
        return self.value
    def __getitem__(self, index):
        return self.value.group(index)

def get(program):
    """Return the text of the 'man' page for the given command-line program.
    """
    text = os.popen('man %s | col -b' % program).readlines()
    return ''.join(text)

class Option:
    def __init__(self, header='', doc=''):
        self.header = header
        self.doc = ''
        self.append(doc)
        # Get option name from header
        self.option = header.split()[0].lstrip('-')

    def getopt(self):
        """Get option name(s) and argument type from option header."""
        text = self.header
        option = ''
        arg = ''
        match = Matcher()
        # Typical option header styles
        # -f, --foobar
        if match.search('(-\w), --[-\w]+', text):
            option = match[1]
        # -f ARG, --foobar=ARG
        elif match.search('(-\w) (\w+), --[-\w]+=\w+', text):
            option = match[1]
            arg = match[2]
        # -foo
        elif match.search('(-[-\w]+)', text):
            option = match[1]
        # -foo ARG
        elif match.search('(-[-\w]+) \[?(\w+)]?', text):
            option = match[1]
            arg = match[2]

        return (option, arg)

    def append(self, text):
        """Append text to the documentation, with extra whitespace removed.
        """
        text = text.replace('\t', ' ')
        text = text.replace('  ', ' ')
        text = text.strip()
        # Join hyphenated words at end of lines
        if self.doc.endswith('-'):
            self.doc = self.doc.rstrip('-') + text
        else:
            self.doc += ' ' + text

    def __str__(self):
        text = self.header + '\n'
        text += textwrap.fill(self.doc.strip())
        return text

def parse(text):
    """Generate meta-GUI from manpage documentation"""
    options = []
    # Match lines defining options
    option = re.compile("^ *-(\w+) ?(.*)")
    for line in text.splitlines():
        if re.match(option, line):
            options.append(Option(line.strip()))
        elif len(options) > 0:
            options[-1].append(line)
    return options




# odict.py

"""Ordered dictionary class, from a recipe_ in the Python Cookbook.

.. _recipe: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/107747
"""

__all__ = [
    'Odict',
    'convert_list',
]

# Python < 3.x
try:
    from UserDict import UserDict
# Python 3.x
except ImportError:
    from collections import UserDict

class Odict (UserDict):
    """Ordered dictionary class, compatible with the builtin `dict`.
    The order in which items are added to the Odict is preserved.
    """
    def __init__(self, keys=None, values=None):
        """Create an Odict from the given keys and values.
        """
        keys = list(keys or [])
        values = list(values or [])
        if len(keys) != len(values):
            raise ValueError("keys and values must be the same length")
        self._keys = keys
        UserDict.__init__(self, dict(zip(keys, values)))

    def __delitem__(self, key):
        """Delete the value from Odict[``key``].
        """
        UserDict.__delitem__(self, key)
        self._keys.remove(key)

    def __setitem__(self, key, item):
        """Set Odict[``key``] = ``item``.
        """
        UserDict.__setitem__(self, key, item)
        if key not in self._keys:
            self._keys.append(key)

    def clear(self):
        """Remove all items from the Odict.
        """
        UserDict.clear(self)
        self._keys = []

    def copy(self):
        """Return an exact copy of the Odict.
        """
        dict_copy = UserDict.copy(self)
        dict_copy._keys = self._keys[:]
        return dict_copy

    def items(self):
        """Return a list of ``(key, value)`` pairs, in order.
        """
        return zip(self._keys, self.values())

    def keys(self):
        """Return the list of keys, in order.
        """
        return self._keys

    def values(self):
        """Return the list of values, in order.
        """
        return [self.get(key) for key in self._keys]

    def popitem(self):
        """Pop the last ``(key, value)`` pair from the Odict, and return it.
        """
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError('dictionary is empty')
        else:
            return self.pop(key)

    def pop(self, key):
        """Pop the item with the given key and return it.
        """
        value = self[key]
        del self[key]
        return value

    def setdefault(self, key, failobj=None):
        """If Odict[key] doesn't exist, set Odict[key] = failobj,
        and return the resulting value of Odict[key].
        """
        result = UserDict.setdefault(self, key, failobj)
        if key not in self._keys:
            self._keys.append(key)
        return result

    def update(self, other_dict, **kwargs):
        """Update the Odict with values from another `dict`.
        """
        UserDict.update(self, other_dict, **kwargs)
        for key in other_dict.keys():
            if key not in self._keys:
                self._keys.append(key)

    def __str__(self):
        """Return a string representation of the Odict.
        """
        lines = []
        for key, value in self.items():
            lines.append("%s: %s" % (key, value))
        return '\n'.join(lines)


def convert_list(choices):
    """Convert a list of choices to an `Odict` (ordered dictionary).
    ``choices`` may be in one of several formats:

        string
            ``'one|two|three'``
        list
            ``['one', 'two', 'three']``
        dict
            ``{'a': "Choice A", 'b': "Choice B"}``
        list-of-lists
            ``[['a', "Choice A"], ['b', "Choice B"], ..]``

    Note: the `dict` form does not preserve order. Use list-of-lists
    to maintain the specified order.
    """
    if type(choices) not in [str, list, dict]:
        raise TypeError("choices must be a string, list, or dictionary.")

    if type(choices) == str:
        choices = choices.split('|')
        return Odict(choices, choices)

    if type(choices) == dict:
        return Odict(choices.keys(), choices.values())

    # choices is a list, but what kind?
    first = choices[0]
    # list of strings
    if type(first) == str:
        return Odict(choices, choices)
    # list of 2-element string lists
    elif type(first) == list and len(first) == 2:
        choices, values = zip(*choices)
        return Odict(choices, values)
    else:
        raise TypeError("choices lists must either be"\
            "['a', 'b', 'c'] or [['a', 'A'], ['b', 'B']] style.")




"""This module defines a Control class and several derivatives. A Control is a
special-purpose GUI widget for setting a value such as a number or filename.
"""

__all__ = [
    'Control',
    # Subclasses
    'Choice',
    'Color',
    'Filename',
    'Flag',
    'FlagOpt',
    'Font',
    'List',
    'ListToOne',
    'ListToMany',
    'Number',
    'SpacedText',
    'Text',
]
import shlex
import re

# Python < 2.5 (I'm looking at YOU, CentOS)
try:
    any
except NameError:
    def any(iterable):
        """Return True if bool(x) is True for any x in iterable."""
        for item in iterable:
            if item:
                return True
        return False

             

# Python < 3.x
try:
    import Tkinter as tk
    from tkFileDialog import \
        (asksaveasfilename, askopenfilename, askopenfilenames)
    from tkColorChooser import askcolor
# Python 3.x
except ImportError:
    import tkinter as tk
    from tkinter.filedialog import \
        (asksaveasfilename, askopenfilename, askopenfilenames)
    from tkinter.colorchooser import askcolor

# Python 3.x
from libtovid import basestring

from libtovid.metagui.widget import Widget
from libtovid.metagui.variable import VAR_TYPES, ListVar
from libtovid.metagui.support import \
    (DragList, ScrollList, FontChooser, PopupScale, ensure_type, ComboBox)
# Used in Control
from libtovid.metagui.tooltip import ToolTip
# Used in Choice control
from libtovid.odict import convert_list


# ---------------
# Exceptions
# ---------------

class NotDrawn (Exception):
    """Exception raised when a Control has not been drawn yet.
    """
    pass


class NoSuchControl (ValueError):
    """Exception raised when a nonexistent Control is referenced.
    """
    pass


# ---------------
# Classes
# ---------------

class Control (Widget):
    """A specialized GUI widget that controls a command-line option.

    Control subclasses may have any number of sub-widgets such as labels,
    buttons or entry boxes; one of the sub-widgets should be linked to
    self.variable like so::

        textbox = tk.Entry(self, textvariable=self.variable)

    See the Control subclasses below for examples of how self.variable,
    get() and set() are used.
    """

    # Dict of all instantiated Controls, indexed by option string
    all = {}

    @staticmethod
    def by_option(option):
        """Return the Control instance for a given option string,
        or ``None`` if no Control has that option string.
        """
        if option != '' and option in Control.all:
            return Control.all[option]
        else:
            raise NoSuchControl("No Control exists for option: '%s'" % option)


    def __init__(self,
                 vartype=str,
                 label='',
                 option='',
                 default='',
                 help='',
                 toggles=False,
                 labelside='left',
                 **kwargs):
        """Create a Control for an option.

        Arguments:

            vartype
                Python type of stored variable
                (str, bool, int, float, list, dict)
            label
                Label shown in the GUI for the Control
            option
                Command-line option associated with this Control,
                or empty to create a positional argument
            default
                Default value for the Control
            help
                Help text to show in a tooltip
            toggles
                Control widget may be toggled on/off
            labelside
                Position of label ('left' or 'top'). It's up to the
                derived Control class to use this appropriately.
            kwargs
                Keyword arguments of the form ``key1=arg1, key2=arg2``

        """
        Widget.__init__(self, label, **kwargs)
        self.vartype = vartype
        self.variable = None
        self.label = label
        self.option = option
        self.default = default or vartype()
        self.help = help
        self.toggles = toggles
        self.labelside = labelside
        self.callbacks = []
        self._callback_name = ''

        # Defined in draw()
        self.checked = None
        self.tooltip = None
        self.check = None

        # Add self to all
        if self.option != '':
            # TODO: Handle multiple Controls with same option
            # (for now, just ignore all but the first one)
            if self.option in Control.all:
                pass
            else:
                Control.all[self.option] = self


    def draw(self, master):
        """Draw the control widgets in the given master.

        Override this method in derived classes, and call the base
        class method::

            Control.draw(self, master)

        """
        Widget.draw(self, master)
        # Create tk.Variable to store Control's value
        if self.vartype in VAR_TYPES:
            self.variable = VAR_TYPES[self.vartype](self)
        else:
            self.variable = tk.Variable(self)
        # Set a trace callback on the variable
        self._add_trace(self.variable)
        # Set default value
        if self.default:
            self.variable.set(self.default)
        # Draw tooltip
        if self.help != '':
            self.tooltip = ToolTip(self, text=self.help, delay=1000)
        # Draw the toggle checkbox if desired
        if self.toggles:
            self.checked = tk.BooleanVar()
            self.check = tk.Checkbutton(self, text='', command=self.toggle,
                                        var=self.checked)
            self.check.pack(side='left')


    def toggle(self):
        """Enable or disable the Control when self.check is toggled.
        """
        # If checkbox is checked, enable Control
        if self.checked.get():
            self.enable()
        # Otherwise, disable (then re-enable checkbox)
        else:
            self.disable()
            self.check.config(state='normal')


    def post(self):
        """Post-draw initialization.
        """
        if not self.enabled:
            self.disable()
        if self.toggles:
            self.toggle()


    def get(self):
        """Return the value of the Control's variable.
        """
        # self.variable isn't defined until draw() is called
        if not self.variable:
            raise NotDrawn("Can't get() from '%s'" % self.name)
        # In some strange cases (like a Number control with an empty Entry)
        # the get() method can raise a ValueError. If so, just return the
        # control's default value.
        try:
            return self.variable.get()
        except ValueError:
            return self.default


    def set(self, value):
        """Set the Control's variable to the given value.
        """
        # self.variable isn't defined until draw() is called
        if not self.variable:
            raise NotDrawn("Must call draw() before set()")
        self.variable.set(value)
        # Set a trace callback on the variable
        self._add_trace(self.variable)


    def set_variable(self, variable):
        """Set the Control to reference the given Tkinter variable.
        """
        need_type = VAR_TYPES[self.vartype]
        if not isinstance(variable, need_type):
            raise TypeError("Variable must be of type '%s'" % need_type)
        self.variable = variable


    def reset(self):
        """Reset the Control's value to the default.
        """
        if self.variable:
            self.set(self.default)


    def focus(self):
        """Set the focus to this Control.
        Override in subclasses if needed.
        """
        pass


    def get_args(self):
        """Return a list of arguments for passing this command-line option.
        `draw` must be called before this function.
        """
        args = []
        value = self.get()

        # Return empty if the control is toggled off
        if not self.enabled:
            return []

        # Skip if unmodified or empty
        elif value == self.default or value == []:
            return []

        # Add option string
        if self.option != '':
            # For <option>, don't pass the option string
            if self.option.startswith('<') and self.option.endswith('>'):
                pass
            # For lists, only pass the option if the list has content
            elif type(value) == list:
                if any(value):
                    args.append(self.option)
            # All others--just append the option string
            else:
                args.append(self.option)

        # List of arguments
        if type(value) == list:
            args.extend(value)
        # Single argument
        else:
            args.append(value)
        return args


    def set_args(self, args):
        """Set control options from the given list of command-line arguments,
        and remove any successfully parsed options and arguments from ``args``.
        """
        # If this control's option is not in args, there's nothing to do
        if self.option not in args:
            return
        # Get the index where the option appears
        index = args.index(self.option)
        # TODO


    def add_callback(self, callback):
        """Add a callback to this Control, which will be called anytime the
        Control's value is modified. The callback will be called with a single
        argument: this Control instance. Callbacks will only be added once.
        """
        # Ensure the callback is callable
        if not callable(callback):
            raise TypeError("Control '%s' callback is not callable." % \
                            self.option)
        # Don't add the same callback more than once
        if callback in self.callbacks:
            return
        self.callbacks.append(callback)


    def _add_trace(self, variable):
        """Add a trace callback to self.variable, and store the callback name.
        """
        # If self.variable is not set yet, do nothing
        if not variable:
            return
        # If this variable doesn't already call self._callback,
        # add it using trace_variable
        found = False
        for mode, callback_name in variable.trace_vinfo():
            if callback_name == self._callback_name:
                found = True
        if not found:
            self._callback_name = variable.trace_variable('w', self._callback)


    def _callback(self, name, index, mode):
        """Callback wrapper, called when variable's value is modified.
        This method calls all other callbacks with the updated value.
        """
        value = self.get()
        for callback in self.callbacks:
            if callable(callback):
                callback(self)


class Choice (Control):
    """Multiple-choice selector, with radiobutton or dropdown style.
    """
    def __init__(self,
                 label="Choices",
                 option='',
                 default='',
                 help='',
                 choices='A|B',
                 style='radio',
                 side='left',
                 **kwargs):
        """Initialize Choice widget with the given label and list of choices.

            label
                Text label for the choices
            option
                Command-line option to set
            default
                Default choice, or empty to use first choice in list
            help
                Help text to show in a tooltip
            choices
                Available choices, in string form: ``one|two|three``, list
                form ``['one', 'two', 'three']``, or as a list-of-lists
                ``[['a', "Use A"], ['b', "Use B"], ..]``.  A dictionary is also
                allowed, as long as you don't care about preserving order.
            style
                ``radio`` for radiobuttons, ``dropdown`` for a drop-down list
            side
                ``left`` for horizontal, ``top`` for vertical arrangement

        """
        self.choices = convert_list(choices)
        Control.__init__(self, str, label, option,
                         default or self.choices.values()[0],
                         help, **kwargs)
        if style not in ['radio', 'dropdown']:
            raise ValueError("Choice style must be 'radio' or 'dropdown'")
        self.style = style
        self.side = side
        # Defined in draw()
        self.combo = None
        self.rb = None


    def draw(self, master):
        """Draw control widgets in the given master.
        """
        Control.draw(self, master)
        if self.style == 'radio':
            frame = tk.LabelFrame(self, text=self.label)
            frame.pack(anchor='nw', fill='x')
            self.rb = {}
            for choice, label in self.choices.items():
                self.rb[choice] = tk.Radiobutton(frame,
                    text=label, value=choice, variable=self.variable)
                self.rb[choice].pack(anchor='nw', side=self.side)
        else: # dropdown/combobox
            label = tk.Label(self, text=self.label)
            label.pack(anchor='w', side=self.labelside)
            self.combo = ComboBox(self, self.choices.keys(),
                                  variable=self.variable)
            self.combo.pack(side='left')
        Control.post(self)


class Color (Control):
    """A color chooser that may have '#RRGGBB' or 'ColorName' values.
    """
    def __init__(self,
                 label="Color",
                 option='',
                 default='',
                 help='',
                 **kwargs):
        """Create a widget that opens a color-chooser dialog.

            label
                Text label describing the color to be selected
            option
                Command-line option to set
            default
                Default color (named color or hexadecimal RGB)
            help
                Help text to show in a tooltip

        """
        Control.__init__(self, str, label, option, default, help, **kwargs)
        # Defined in draw()
        self.button = None
        self.editbox = None


    def draw(self, master):
        """Draw control widgets in the given master.
        """
        Control.draw(self, master)
        label = tk.Label(self, text=self.label)
        label.pack(side=self.labelside)
        # Button for opening a color picker popup
        self.button = tk.Button(self, text='Color', command=self.pick_color)
        self.button.pack(side='left')
        # Textbox for typing in an RGB hex value or color name
        self.editbox = tk.Entry(self, textvariable=self.variable, width=8)
        self.editbox.pack(side='left', fill='y')
        # Update the color preview when the variable changes
        self.add_callback(self.update_color)
        # Indicate the current (default) color
        if self._is_hex_rgb(self.default):
            self.indicate_color(self.default)
        Control.post(self)


    def update_color(self, event=None):
        """Event handler to update the color preview.
        """
        color = self.variable.get().strip()
        try:
            self.set(color)
        except (ValueError):
            pass


    def pick_color(self):
        """Event handler for the color picker button; choose and set a color.
        """
        current = self.hexcolor(self.get())
        rgb, color = askcolor(current)
        if color:
            self.set(str(color))


    def set(self, color):
        """Set the current color to an RGB hex value or color name.
        """
        # Update variable with whatever color name or RGB value was given
        # (even if it's not necessarily a valid color)
        self.variable.set(color)

        # Show the color in the indicator button
        self.indicate_color(self.hexcolor(color))


    def hexcolor(self, color):
        """Return an 8-bit '#RRGGBB' hex string for the given color.
        If a given color name is unknown, return '#FFFFFF' (white).
        """
        # If color is already hex, return it
        if self._is_hex_rgb(color):
            return color

        # Try to get color by name, converting from 16-bit to 8-bit RGB
        try:
            rgb = [x / 256 for x in self.winfo_rgb(color)]
        # Use white for any unknown color name
        except (tk.TclError):
            rgb = [255, 255, 255]

        return self._rgb_to_hex(rgb)


    def indicate_color(self, bg_color):
        """Change the button background color to the given RGB hex value.
        """
        if not self._is_hex_rgb(bg_color):
            raise ValueError("indicate_color needs an 8-bit #RRGGBB hex string")
        # Choose a foreground color that will be visible
        r, g, b = self._hex_to_rgb(bg_color)
        if (r + g + b) > 384: # hack
            fg_color = '#000000' # black
        else:
            fg_color = '#ffffff' # white
        # Set button background color to chosen color
        self.button.config(background=bg_color, foreground=fg_color)


    # Static methods for supporting functions
    @staticmethod
    def _is_hex_rgb(color):
        """Return True if color appears to be a hex '#RRGGBB' value.
        Both three-digit and six-digit hex codes are allowed.

        Examples::

            >>> _is_hex_rgb('#FF0080')
            True
            >>> _is_hex_rgb('#F08')
            True
            >>> _is_hex_rgb('#FF008')
            False

        """
        if not isinstance(color, basestring):
            return False
        elif re.match('^#[0-9a-fA-F]{6}$', color):
            return True
        elif re.match('^#[0-9a-fA-F]{3}$', color):
            return True
        else:
            return False


    @staticmethod
    def _hex_to_rgb(color):
        """Convert a hexadecimal color string '#RRGGBB' to an RGB tuple.

        Example::

            >>> _hex_to_rgb('#FF0080')
            (255, 0, 128)

        """
        color = color.lstrip('#')
        red, green, blue = (color[0:2], color[2:4], color[4:6])
        return (int(red, 16), int(green, 16), int(blue, 16))


    @staticmethod
    def _rgb_to_hex(rgb_tuple):
        """Convert an RGB tuple into a hexadecimal color string '#RRGGBB'.

        Example::

            >>> _rgb_to_hex((255, 0, 128))
            '#FF0080'

        """
        red, green, blue = rgb_tuple
        return '#%02x%02x%02x' % (red, green, blue)


class Filename (Control):
    """Filename entry box with browse button.
    """
    def __init__(self,
                 label='Filename',
                 option='',
                 default='',
                 help='',
                 action='load',
                 desc='Select a file to load',
                 filetypes='all',
                 **kwargs):
        """Create a Filename with label, text entry, and browse button.

            label
                Text of label next to file entry box
            option
                Command-line option to set
            default
                Default filename
            help
                Help text to show in a tooltip
            action
                Do you intend to 'load' or 'save' this file?
            desc
                Brief description (shown in title bar of file browser dialog)
            filetypes
                Types of files to show in the file browser dialog. May be 'all'
                for all file types, or a list of ``('label', '*.ext')`` tuples.

        """
        Control.__init__(self, str, label, option, default, help, **kwargs)
        self.action = action
        self.desc = desc
        if filetypes == 'all':
            self.filetypes = [('All Files', '*.*')]
        else:
            self.filetypes = filetypes
        # Defined by draw()
        self.entry = None
        self.button = None


    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        # Create and pack widgets
        label = tk.Label(self, text=self.label, justify='left')
        label.pack(side=self.labelside)
        self.entry = tk.Entry(self, textvariable=self.variable)
        self.button = tk.Button(self, text="Browse...", command=self.browse)
        self.entry.pack(side='left', fill='x', expand=True)
        self.button.pack(side='left')
        Control.post(self)


    def browse(self, event=None):
        """Event handler when browse button is pressed"""
        if self.action == 'save':
            filename = asksaveasfilename(parent=self, title=self.desc)
        else: # 'load'
            filename = askopenfilename(parent=self, title=self.desc,
                                       filetypes=self.filetypes)
        # Got a filename? Display it
        if filename:
            self.set(filename)


class Flag (Control):
    """Yes/no checkbox, for flag-type options.
    """
    def __init__(self,
                 label="Flag",
                 option='',
                 default=False,
                 help='',
                 enables=None,
                 **kwargs):
        """Create a Flag widget with the given label and default value.

            label
                Text label for the flag
            option
                Command-line flag passed
            default
                Default value (``True`` or ``False``)
            help
                Help text to show in a tooltip
            enables
                Option or list of options to enable when the Flag is checked

        """
        Control.__init__(self, bool, label, option, default, help, **kwargs)

        # Ensure the "enables" arg is the right type
        if not enables:
            self.enables = []
        elif isinstance(enables, list):
            self.enables = enables
        elif isinstance(enables, basestring):
            self.enables = [enables]
        else:
            raise TypeError("Flag 'enables' argument must be"
                            " an option string or list of option strings"
                            " (got %s instead)" % enables)
        # Will be a list of enabled Controls, filled in by draw()
        self.controls = []


    def draw(self, master):
        """Draw the Flag in the given master widget.
        """
        Control.draw(self, master)
        self.check = tk.Checkbutton(self, text=self.label,
                                    variable=self.variable,
                                    command=self.enabler)
        self.check.pack(side=self.labelside)

        # Enable/disable related controls
        self.controls = [Control.by_option(opt) for opt in self.enables]
        Flag.enabler(self)
        Control.post(self)


    def enabler(self):
        """Enable/disable related Controls based on Flag state.
        """
        for control in self.controls:
            if self.get():
                if control.is_drawn:
                    control.enable()
                else:
                    control.enabled = True
            else:
                if control.is_drawn:
                    control.disable()
                else:
                    control.enabled = False


    def get_args(self):
        """Return a list of arguments for passing this command-line option.
        `draw` must be called before this function.
        """
        args = []
        # If flag is true, and option is nonempty, append to args
        if self.get() and self.option:
            args.append(self.option)
        return args


class FlagOpt (Flag):
    """Like a normal Flag, but has an optional argument following it.
    """
    def __init__(self,
                 label="Flag",
                 option='',
                 default=False,
                 help='',
                 control=None,
                 **kwargs):
        """Create a FlagOpt widget; like a Flag, but has an optional argument
        which may be set using an associated Control.

            label
                Text label for the flag
            option
                Command-line flag passed
            default
                Default value (True or False)
            help
                Help text to show in a tooltip
            control
                Another Control, for setting the argument value (required)

        """
        Flag.__init__(self, label, option, default, help, **kwargs)
        if control != None:
            ensure_type("FlagOpt needs a Control instance", Control, control)
        self.control = control


    def draw(self, master):
        """Draw the Flag and associated Control in the given master.
        """
        Flag.draw(self, master)
        # Pack the arg control next to the flag checkbox
        self.control.draw(self)
        self.control.pack(anchor='nw', side='left', fill='x', expand=True)
        # Disable if flag defaults to false
        if not self.default:
            self.control.disable()


    def enabler(self):
        """Enable or disable the arg Control when the flag changes.
        """
        Flag.enabler(self)
        if self.get():
            self.control.enable()
        else:
            self.control.disable()


    def get_args(self):
        """If the flag is enabled, return the flag and associated option.
        """
        args = Flag.get_args(self)
        if len(args) > 0:
            args.extend(self.control.get_args())
        return args


class Font (Control):
    """Font name chooser."""
    def __init__(self,
                 label='Font',
                 option='',
                 default='Helvetica',
                 help='',
                 **kwargs):
        """Create a widget that opens a font chooser dialog.

            label
                Text label for the font
            option
                Command-line option to set
            default
                Default font name
            help
                Help text to show in a tooltip

        """
        Control.__init__(self, str, label, option, default, help, **kwargs)
        # Defined by draw()
        self.button = None


    def draw(self, master):
        """Draw the Font selector in the given master widget.
        """
        Control.draw(self, master)
        tk.Label(self, text=self.label).pack(side=self.labelside)
        self.button = tk.Button(self, textvariable=self.variable,
                                command=self.choose)
        self.button.pack(side='left', padx=8)
        Control.post(self)


    def choose(self):
        """Open a font chooser to select a font.
        """
        chooser = FontChooser(self)
        if chooser.result:
            self.variable.set(chooser.result)


class Number (Control):
    """Numeric entry box or slider.
    """
    def __init__(self,
                 label="Number",
                 option='',
                 default=0,
                 help='',
                 min=1,
                 max=10,
                 units='',
                 style='spin',
                 step=1,
                 **kwargs):
        """Create a number-setting widget.

            label
                Text label describing the meaning of the number
            option
                Command-line option to set
            default
                Default value
            help
                Help text to show in a tooltip
            min, max
                Range of allowable numbers (inclusive)
            units
                Units of measurement (ex. 'kbits/sec'), used as a label
            style
                'spin' for a spinbox,
                'scale' for a slider,
                'popup' for a slider in a popup window
            step
                For 'scale' style, the amount to increment or
                decrement when the slider is moved

        The default/min/max may be integers or floats.
        """
        Control.__init__(self, type(default), label, option, default,
                         help, **kwargs)
        self.min = min
        self.max = max
        self.units = units
        self.style = style
        self.step = step
        # Defined by draw()
        self.number = None


    def draw(self, master):
        """Draw the Number entry in the given master widget.
        """
        Control.draw(self, master)
        tk.Label(self, name='label', text=self.label).pack(side=self.labelside)
        if self.style == 'spin':
            self.number = tk.Spinbox(self, from_=self.min, to=self.max,
                                     width=4, textvariable=self.variable)
            self.number.pack(side='left')
            tk.Label(self, name='units', text=self.units).pack(side='left')

        elif self.style == 'scale':
            tk.Label(self, name='units', text=self.units).pack(side='left')
            self.number = tk.Scale(self, from_=self.min, to=self.max,
                                   resolution=self.step,
                                   tickinterval=(self.max - self.min),
                                   variable=self.variable, orient='horizontal')
            self.number.pack(side='left', fill='x', expand=True)

        else: # 'popup'
            def popup():
                """Show a popup scale for setting the variable's value."""
                scale = PopupScale(self)
                if scale.result is not None:
                    self.variable.set(scale.result)
            tk.Button(self, textvariable=self.variable,
                      command=popup).pack(side='left')
            tk.Label(self, name='units', text=self.units).pack(side='left')

        Control.post(self)


    def enable(self, enabled=True):
        """Enable or disable all sub-widgets.
        """
        # Overridden to make Scale widget look disabled
        Widget.enable(self, enabled)
        if self.style == 'scale':
            if enabled:
                self.number['fg'] = 'black'
                self.number['troughcolor'] = 'white'
            else:
                self.number['fg'] = '#A3A3A3'
                self.number['troughcolor'] = '#D9D9D9'


class Text (Control):
    """Text string entry box.
    """
    def __init__(self,
                 label="Text",
                 option='',
                 default='',
                 help='',
                 width=None,
                 **kwargs):
        """Create a text-entry control.

            label
                Label for the text
            option
                Command-line option to set
            default
                Default value of text widget
            help
                Help text to show in a tooltip
            width
                Width in characters of the text field, or
                ``None`` to maximize width in available space
        """
        Control.__init__(self, str, label, option, default, help, **kwargs)
        self.width = width
        # Defined by draw()
        self.entry = None
        self.parent_list = None


    def draw(self, master):
        """Draw the Text control in the given master widget.
        """
        Control.draw(self, master)
        label = tk.Label(self, text=self.label, justify='left')
        label.pack(side=self.labelside)
        self.entry = tk.Entry(self, textvariable=self.variable)
        if self.width:
            self.entry.config(width=self.width)
            self.entry.pack(side='left')
        else:
            self.entry.pack(side='left', fill='x', expand=True)
        Control.post(self)
        self.entry.bind('<Return>', self.next_item)


    def focus(self):
        """Highlight the text entry box for editing.
        """
        self.entry.select_range(0, 'end')
        self.entry.focus_set()


    def set_parent(self, parent_list):
        """Set the parent List control.
        """
        if not isinstance(parent_list, List):
            raise TypeError("Text control must have a List as its parent.")
        self.parent_list = parent_list


    def next_item(self, event):
        """Select the next item in the parent listbox.
        """
        if self.parent_list is not None:
            self.parent_list.listbox.next_item(event)


class SpacedText (Text):
    """Text string interpreted as a space-separated list of strings
    """
    def __init__(self,
                 label="List",
                 option='',
                 default='',
                 help='',
                 **kwargs):
        Text.__init__(self, label, option, default, help, **kwargs)


    def draw(self, master):
        """Draw SpacedText control in the given master widget.
        """
        Text.draw(self, master)


    def get(self):
        """Get current text, split into a list of strings.
        """
        text = Text.get(self)
        return shlex.split(text)


    def set(self, listvalue):
        """Set the text equal to the given list, joined with spaces and
        double-quoted. Double-quotes in list values are backslash-escaped.
        """
        def quote(val):
            """Put double-quotes around the given value."""
            return '"%s"' % val.replace('"', '\\"')
        text = ' '.join([quote(val) for val in listvalue])
        Text.set(self, text)


class List (Control):
    """A list of values, editable with a given Control.

    Examples::

        List('Filenames', '-files', control=Filename())
        List('Colors', '-colors', control=Color())
        List('Texts', '-texts', control=Text())
        List('Choices', '-choices', control=Choice())

    """
    def __init__(self,
                 label="Text List",
                 option='',
                 default=None,
                 help='',
                 control=Text(),
                 **kwargs):
        """Create a control for a list-type option.
        """
        ensure_type("List requires a Control instance", Control, control)
        Control.__init__(self, list, label, option, default, help, **kwargs)
        self.control = control
        # Defined by draw()
        self.listbox = None


    def draw(self, master):
        """Draw the List and associated Control in the given master.
        """
        Control.draw(self, master)

        # Draw the outer frame, containing the listbox and tool frame
        frame = tk.LabelFrame(self, text=self.label)
        self.listbox = self._draw_listbox(frame)
        tool_frame = self._draw_tool_frame(frame)

        # Pack everything
        frame.pack(fill='both', expand=True)
        self.listbox.pack(fill='both', expand=True)
        tool_frame.pack(fill='x')

        self._bind_control_to_listbox(self.control, self.listbox)


    def _draw_listbox(self, master, allow_add_remove=True):
        """Draw a listbox with appropriate bindings in the given master,
        and return the new listbox.
        """
        # Scrolled or draggable listbox
        if allow_add_remove:
            listbox = DragList(master, self.variable)
        else:
            listbox = ScrollList(master, self.variable)
        listbox.bind('<Return>', listbox.next_item)
        listbox.callback('select', self.select)

        return listbox


    def _draw_tool_frame(self, master, allow_add_remove=True):
        """Draw the child control and add/remove buttons in a new frame with
        the given master, and return the new frame.
        """
        # Frame to hold add/remove buttons and child Control
        tool_frame = tk.Frame(master)

        # Add/remove buttons (only shown if allow_add_remove)
        if allow_add_remove:
            add_button = \
                tk.Button(tool_frame, text="Add", command=self.add)
            remove_button = \
                tk.Button(tool_frame, text="Del", command=self.remove)
            add_button.pack(fill='x', side='left')
            remove_button.pack(fill='x', side='left')

        # Draw associated Control
        self.control.draw(tool_frame)
        self.control.pack(fill='x', side='left', expand=True)

        return tool_frame


    def _bind_control_to_listbox(self, control, listbox):
        """Bind a given control to a listbox.
        """
        # Set up bindings on the control
        control.bind('<Return>', listbox.next_item)
        if isinstance(control, Text):
            control.set_parent(self)
        # Disable the control until values are added
        control.disable()

        # Update selected list item when control's variable is modified
        control.add_callback(self.modify)


    def refresh_control(self):
        """Enable the Control if there are items in the list,
        otherwise disable it.
        """
        if self.listbox.items.count() > 0:
            self.control.enable()
        else:
            self.control.disable()


    def set(self, value_list):
        """Set all list values.
        """
        # Use the listbox's set() method, so the relevant callbacks
        # will be summoned for any child lists
        self.listbox.set(value_list)
        self.refresh_control()


    def modify(self, control):
        """Event handler when the Control's variable is modified.
        """
        index = self.listbox.curindex
        new_value = control.get()
        # Only update the list if the new value is different
        if self.variable[index] != new_value:
            self.variable[index] = new_value


    def select(self, index, value):
        """Select an item in the list and enable editing.
        """
        # Only set control if value is different
        if self.control.get() != value:
            self.control.set(value)
        self.control.focus()


    def add(self):
        """Event handler for the "Add" button.
        """
        # For filenames, show a file chooser to add one or more files
        if isinstance(self.control, Filename):
            files = askopenfilenames(parent=self, title='Add files',
                                     filetypes=self.control.filetypes)
            self.listbox.add(*files)
        # For all others, add an item to the list
        else:
            value = self.control.default
            self.listbox.add(value)

        # Select the last item in the list, and enable the control
        self.listbox.select_index(-1)
        self.control.enable()


    def remove(self):
        """Event handler for the "Remove" button.
        """
        if self.listbox.items.count() > 0:
            index = self.listbox.curindex
            self.listbox.delete(index)
        self.refresh_control()


    def set_variable(self, variable):
        """Set the List to use the given ListVar as its variable.
        """
        Control.set_variable(self, variable)
        self.listbox.set_variable(variable)
        self.refresh_control()


class _SubList (List):
    """Base class for ListToOne and ListToMany.
    """
    def __init__(self,
                 parent,
                 label="_SubList",
                 option='',
                 default=None,
                 help='',
                 control=Text(),
                 filter=lambda x: x,
                 side='left',
                 **kwargs):
        """Create a _SubList. This should only be called from derived classes!
        """
        List.__init__(self, label, option, default, help, control)

        # Check for correct values / types
        if not isinstance(parent, (List, basestring)):
            raise TypeError("Parent must be a List or an option string.")
        if not callable(filter):
            raise TypeError("Translation filter must be a function.")
        if side not in ['left', 'top']:
            raise ValueError("ChildList 'side' must be 'left' or 'top'")

        self.parent = parent
        self.filter = filter
        self.side = side
        # Set by draw()
        self.parent_is_copy = False
        self.parent_listbox = None


    def draw(self, master, allow_add_remove=True):
        """Draw the parent copy and related list Control,
        side by side in the given master.
        """
        # Calling Control.draw here instead of List.draw, since we're
        # replacing all the List.draw functionality
        Control.draw(self, master)

        # Frame to wrap both lists in
        outer_frame = tk.Frame(master, padx=8, pady=8)

        # Draw parent listbox in the outer frame
        parent_frame = self._draw_parent(outer_frame)

        # Draw this list in its own frame
        # Frame to draw listbox and child Control in
        list_frame = tk.LabelFrame(outer_frame, text=self.label)

        # Draw the listbox and tool frame
        self.listbox = self._draw_listbox(list_frame, allow_add_remove)
        tool_frame = self._draw_tool_frame(list_frame, allow_add_remove)
        # Pack the listbox and tool frame
        self.listbox.pack(fill='both', expand=True)
        tool_frame.pack(fill='x')

        self._bind_control_to_listbox(self.control, self.listbox)

        # Pack the parent and the current list
        parent_frame.pack(side=self.side, anchor='nw', fill='both', expand=True)
        list_frame.pack(side=self.side, anchor='nw', fill='both', expand=True)
        # Pack the outer frame containing both lists
        outer_frame.pack(fill='both', expand=True)


    def _draw_parent(self, master):
        """Draw the parent list in the given master, and return the frame
        containing the parent listbox.
        """
        # If parent is a string, look up the parent control by option name
        # and treat the parent as a copy
        if isinstance(self.parent, basestring):
            self.parent_is_copy = True
            try:
                parent_control = Control.by_option(self.parent)
            except NoSuchControl:
                raise
            else:
                self.parent = parent_control
        # Or use the parent Control itself
        else:
            self.parent_is_copy = False

        ensure_type("ChildList parent must be a List", List, self.parent)

        # Draw the read-only copy of parent's values
        if self.parent_is_copy:
            # FIXME: Not great to bury attribute initialization here
            parent_frame = tk.LabelFrame(master, text="%s (copy)" % \
                                         self.parent.label)
            self.parent_listbox = ScrollList(parent_frame, self.parent.variable)
            self.parent_listbox.pack(expand=True, fill='both')
        # Or draw the parent Control itself
        else:
            parent_frame = self.parent
            self.parent.draw(master)
            self.parent_listbox = self.parent.listbox

        return parent_frame


class ListToOne (_SubList):
    """A List with one value for each value in a parent list.

    This is like a regular List, except that it has a parent:child
    relationship of 1:1 (each parent item has a one child item).

    Behavior:

        - Each item in parent list maps to one item in the child list
        - Parent copy and child list scroll in unison
        - If item in child list is selected, parent item is selected also
        - Drag/drop is allowed in parent list only

    Assumptions:

        - If item in parent is selected, child item/list is selected also
        - It item is added to parent, new child item/list is added also
        - If item in parent is deleted, child item/list is deleted also
        - Child option string is only passed once

    """
    def __init__(self,
                 parent,
                 label="ListToOne",
                 option='',
                 default=None,
                 help='',
                 control=Text(),
                 filter=lambda x: x,
                 side='left',
                 **kwargs):
        """Create a list having a 1:1 mapping to another List.

            parent
                Parent List object, or the option string of the parent List
                control declared elsewhere
            filter
                A function that translates parent values into child values
            side
                Pack the parent to the 'left' of child or on 'top' of child
        """
        _SubList.__init__(self, parent, label, option, default, help,
                          control, filter, side, **kwargs)


    def add_callbacks(self):
        """Add callback functions for add/remove in the parent Control.
        """
        def insert(index, value):
            """When a new item is inserted in the parent list,
            insert a corresponding (filtered) item into the child list.
            """
            self.variable.insert(index, self.filter(value))
            self.control.enable()

        def remove(index, value):
            """When an item is removed from the parent list,
            remove the corresponding item from the child list.
            """
            try:
                self.variable.pop(index)
            except IndexError:
                pass
            # Disable child editor control if child list is empty
            if self.listbox.items.count() == 0:
                self.control.disable()

        def swap(index_a, index_b):
            """When two items are swapped in the parent list,
            swap the corresponding items in the child list.
            """
            self.listbox.swap(index_a, index_b)

        def select(index, value):
            """When an item is selected in the parent list,
            select the corresponding item in the child list.
            """
            pass # Already handled by listboxes being linked

        self.parent_listbox.callback('select', select)
        self.parent.listbox.callback('insert', insert)
        self.parent.listbox.callback('remove', remove)
        self.parent.listbox.callback('swap', swap)


    def draw(self, master):
        """Draw the parent copy and related list Control,
        side by side in the given master.
        """
        _SubList.draw(self, master, allow_add_remove=False)
        # 1:1, parent listbox is linked to this one
        self.parent_listbox.link(self.listbox)
        # Add callbacks to handle changes in parent
        self.add_callbacks()


    def get_args(self):
        """Return a list of arguments for the contained list(s).
        """
        args = []
        # Add parent args, if parent was defined here
        if not self.parent_is_copy:
            args.extend(self.parent.get_args())
        # Add child args
        args.extend(List.get_args(self))
        # Return args only if some list items are non-empty
        if any(args):
            return args
        else:
            return []


class ListToMany (_SubList):
    """A List with many values for each value in a parent list.

    This is like a regular List, except that it has a parent:child
    relationship of 1:* (each parent item has a list of child items).

    Behavior:

        - Each item in parent list maps to a list of items in the child list
        - Parent copy and child list scroll independently
        - If item in child is selected, parent is unaffected
        - Drag/drop is allowed in the child Control

    Assumptions:

        - If item in parent is selected, child item/list is selected also
        - It item is added to parent, new child item/list is added also
        - If item in parent is deleted, child item/list is deleted also
        - Child option string is only passed once

    """
    def __init__(self,
                 parent,
                 label="ListToMany",
                 option='',
                 default=None,
                 help='',
                 control=Text(),
                 filter=lambda x: x,
                 side='left',
                 **kwargs):
        """Create a list having a 1:* mapping to another List.

            parent
                Parent List object, or the option string of the parent List
                control declared elsewhere
            filter
                A function that translates parent values into child values
            side
                Pack the parent to the 'left' of child or on 'top' of child

        Keyword arguments:

            index
                True to pass an additional argument between the child list's
                option and arguments with the 1-based index of the child list.
        """
        _SubList.__init__(self, parent, label, option, default, help,
                          control, filter, side, **kwargs)
        # Will hold a list of ListVars
        self.listvars = {}
        # Handle keyword args
        self.index = kwargs.get('index', False)
        # Hack to support adding multiple lists in increasing order
        self.curindex = 0


    def draw(self, master):
        """Draw the parent copy and related list Control,
        side by side in the given master.
        """
        _SubList.draw(self, master, allow_add_remove=True)
        # Add callbacks to handle changes in parent
        self.add_callbacks()


    def add_callbacks(self):
        """Add callback functions for add/remove in the parent Control.
        """
        def insert(index, value):
            """When a new item is inserted in the parent list,
            insert a new child list (initially empty) for that item.
            """
            self.listvars[index] = ListVar(self)

        def remove(index, value):
            """When an item is removed from the parent list,
            remove the corresponding child list.
            """
            del self.listvars[index]

        def swap(index_a, index_b):
            """When two items are swapped in the parent list,
            swap the two corresponding child lists.
            """
            a_var = self.listvars[index_a]
            self.listvars[index_a] = self.listvars[index_b]
            self.listvars[index_b] = a_var

        def select(index, value):
            """When an item is selected in the parent list,
            display the corresponding child list for editing.
            """
            listvar = self.listvars[index]
            self.set_variable(listvar)

        self.parent_listbox.callback('select', select)
        self.parent.listbox.callback('insert', insert)
        self.parent.listbox.callback('remove', remove)
        self.parent.listbox.callback('swap', swap)


    def get_args(self):
        """Return a list of arguments for the contained list(s).
        """
        args = []
        # Add parent args, if parent was defined here
        if not self.parent_is_copy:
            args.extend(self.parent.get_args())
        # Append arguments for each child list
        for index, list_var in self.listvars.items():
            self.set_variable(list_var)
            child_args = List.get_args(self)
            if child_args:
                child_option = child_args.pop(0)
                args.append(child_option)
                if self.index:
                    args.append(index+1)
                args.extend(child_args)
        # Return args only if some list items are non-empty
        if any(args):
            return args
        else:
            return []


    def set(self, items):
        """Append the given items to listvars.
        """
        # FIXME: This is an abuse of the set() function
        # If an index is prepended, insert at that index
        if self.index:
            index = int(items.pop(0)) - 1
        # Otherwise, set at indices in increasing order
        else:
            index = self.curindex
            self.curindex += 1

        self.listvars[index] = ListVar(self, items)


    def reset(self):
        """Reset the list values to empty.
        """
        self.curindex = 0
        self.listvars = {}


# Exported control classes, indexed by name
CONTROLS = {
    'Choice': Choice,
    'Color': Color,
    'Filename': Filename,
    'Flag': Flag,
    'FlagOpt': FlagOpt,
    'Font': Font,
    'Number': Number,
    'Text': Text,
    'SpacedText': SpacedText,
    'List': List,
}

# Demo
if __name__ == '__main__':
    root = tk.Tk()
    for name, control in CONTROLS.items():
        frame = tk.LabelFrame(root, text=name, padx=10, pady=10,
                          font=('Helvetica', 10, 'bold'))
        frame.pack(fill='both', expand=True)
        widget = control()
        widget.draw(frame)
        widget.pack(fill='both')
    root.mainloop()




"""Logging utilities for metagui.
"""

import inspect
import sys
import os
import linecache

def dump_args(func):
    """Decorator for logging calls to the function, along with
    the arguments passed to the function.
    """
    argnames = func.func_code.co_varnames[:func.func_code.co_argcount]

    def func_wrapper(*args, **kw):
        all_args = zip(argnames, args) + kw.items()
        arg_string = ', '.join(['%s=%r' % (k, v) for k, v in all_args])
        print("Calling %s(%s)" % (func.__name__, arg_string))
        return func(*args, **kw)

    return func_wrapper


def trace(f):
    """Trace lines in the function.
    From http://wiki.python.org/moin/PythonDecoratorLibrary
    """
    def globaltrace(frame, why, arg):
        if why == "call":
            return localtrace
        return None

    def localtrace(frame, why, arg):
        if why == "line":
            # record the file name and line number of every trace
            filename = frame.f_code.co_filename
            lineno = frame.f_lineno

            bname = os.path.basename(filename)
            print("%s(%d): %s" % \
                (bname, lineno, linecache.getline(filename, lineno)))
        return localtrace

    def _f(*args, **kwds):
        sys.settrace(globaltrace)
        result = f(*args, **kwds)
        sys.settrace(None)
        return result

    return _f



#! /usr/bin/env python
# setup.py

"""Distutils installation script for tovid.

To install, run:

    sudo ./setup.py install

or:

    su -c './setup.py install'

At this time, there is no 'uninstall' mechanism...
"""


def svn_version():
    """Return the current SVN revision number, as reported by 'svn info',
    as a string like 'svn-r1234'. If svn is not installed, or if something goes
    wrong, return 'svn-unknown'
    """
    try:
        from commands import getoutput
    except ImportError:
        # python 3
        from subprocess import getoutput

    rev_line = getoutput('svn info 2>/dev/null | grep ^Revision')
    # If rev_line is found, get the revision number
    if rev_line:
        return 'svn-r' + rev_line.split(':')[1].strip()
    # If rev_line is empty, try using bzr instead
    else:
        rev_line = getoutput('bzr log -l 1 2>/dev/null | grep "^svn revno"')
        # "svn revno: NNNN (on /trunk)"
        if rev_line:
            return 'svn-r' + rev_line.split(':')[1].strip().split(' ')[0]
    # If all else fails...
    return 'svn-unknown'

# Current version number of tovid, as a string.
# Examples:
# Current SVN version number
#_tovid_version = svn_version()
# Official release number
_tovid_version = '0.35.2'


import os
import sys
import shutil
from distutils.core import setup, Command
from distutils.command.install import install
from distutils.command.build import build
from distutils.sysconfig import get_config_var


class InstallCommand (install):
    """Extended install command--automatically records installed files
    to $prefix/lib/tovid/.install.log, so they can be uninstalled later.
    """
    def initialize_options(self):
        install.initialize_options(self)
        # Record installation log to a temporary file
        self.record = 'install.log'

    def run(self):
        """Install tovid, and write a list of installed files to
        $prefix/lib/tovid/.install.log.
        """
        install.run(self)
        # For whatever reason, on Ubuntu, self.prefix only points to /usr/local
        # if setup.py was explicitly called with the --prefix option.
        # self.install_data does correctly point to /usr/local on Ubuntu
        # (both with and without the --prefix option)
        prefix = self.install_data
        print("Installing tovid to %s" % prefix)
        # Move temporary install log to $prefix/lib/tovid
        install_log = os.path.join(prefix, 'lib', 'tovid', '.install.log')
        print("Moving install.log to '%s'" % install_log)
        shutil.move(self.record, install_log)


class UninstallCommand (Command):
    description = "remove everything that was installed by the install command"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def get_prefix(self):
        """Return the prefix (/usr or /usr/local) where tovid is installed,
        or '' if the prefix could not be determined.
        """
        # Determine which tovid is in the $PATH
        tovid = os.popen('which tovid').read()
        # Get $prefix/bin
        bin_prefix = os.path.dirname(tovid)
        # Get $prefix
        prefix = os.path.dirname(bin_prefix)
        return prefix

    def run(self):
        """Uninstall tovid, removing all files listed in
        $prefix/lib/tovid/.install.log.
        """
        # Determine the installed prefix
        prefix = self.get_prefix()
        if not prefix:
            print("Could not find tovid in your $PATH. "
                  "You may need to uninstall manually")
            return
        print("Uninstalling tovid from %s" % prefix)
        install_log = os.path.join(prefix, 'lib', 'tovid', '.install.log')
        # If install log doesn't exist, there's nothing to do
        if not os.path.exists(install_log):
            print("Missing installation log: '%s'" % install_log)
            print("Either tovid is not installed, or "
                  "something went wrong during installation. "
                  "You may need to uninstall manually.")
            return
        # Remove each file found in install log
        for line in open(install_log, 'r'):
            filename = line.strip(' \n')
            print("Removing '%s'" % filename)
            os.remove(filename)
        # Remove the install log itself
        os.remove(install_log)
        # Remove the $prefix/lib/tovid directory
        os.removedirs(os.path.dirname(install_log))


class BuildDocs (Command):
    description = "build the tovid documentation (currently just the manpage)"
    user_options = []

    def initialize_options(self):
        self.source = os.path.join('docs', 'src', 'en', 'tovid.t2t')
        self.target = os.path.join('docs', 'man', 'tovid.1')

    def finalize_options(self):
        pass

    def run(self):
        """Build the tovid manual page.
        """
        # For python 3 txt2tags needs an unofficial script
        import sys
        if (sys.version_info > (3, 0)):
            txt2tags = 'docs/scripts/txt2tags'
        else:
            txt2tags = 'txt2tags'

        # Build only if target does not exist, or if source is newer than target
        mod = os.path.getmtime
        if not os.path.exists(self.target) or (mod(self.source) > mod(self.target)):
            print("Rebuilding tovid manual page")
            command = '%s -t man -i "%s" -o "%s"' % (txt2tags, self.source, self.target)
            os.system(command)
        else:
            print("Manual page already built, not building again")


class BuildTovidInit (Command):
    description = "build tovid-init"
    user_options = []

    def initialize_options(self):
        self.source = os.path.join('src', 'tovid-init.in')
        self.target = os.path.join('src', 'tovid-init')
        # Touch the source file to ensure that it gets rebuilt
        os.utime(self.source, None)

    def finalize_options(self):
        pass

    def run(self):
        """Build src/tovid-init from tovid-init.in.
        """
        # We basically just need to replace @VERSION@ in tovid-init.in with
        # the current version of tovid.
        lines = [line.replace('@VERSION@', _tovid_version)
                 for line in open(self.source, 'r')]
        # Write all lines to the target file
        outfile = open(self.target, 'w')
        outfile.writelines(lines)
        outfile.close()


# Build tovid-init with regular 'build' command
build.sub_commands.append(('build_tovid_init', None))
build.sub_commands.append(('build_docs', None))

# The actual setup
setup(
    name = 'tovid',
    version = _tovid_version, # Defined at the top of this file
    url = 'http://tovid.wikia.com/',
    license = 'GPL',
    maintainer = 'grepper',
    maintainer_email = 'grepper@gmail.com',
    description = 'A suite of tools for making video DVDs',
    platforms = 'Linux',
    download_url = 'https://github.com/tovid-suite/tovid/',
    long_description = 'A suite of tools for making video DVDs',

    cmdclass = {
        'install': InstallCommand,
        'uninstall': UninstallCommand,
        'build_docs': BuildDocs,
        'build_tovid_init': BuildTovidInit,
    },

    # Python modules go into /$PREFIX/lib/pythonX.Y/(site|dist)-packages
    packages = [
        'libtovid',
        'libtovid.guis',
        'libtovid.util',
        'libtovid.metagui',
    ],

    # Executable files go into /$PREFIX/bin/
    scripts = [
        # Main executable
        'src/tovid',

    ],

    # Extra files installed relative to /$PREFIX/
    data_files = [
        ('lib/tovid', [
            # Bash scripts
            'src/idvid',
            'src/makedvd',
            'src/todisc',
            'src/todisc-fade-routine',
            'src/makempg',
            'src/tovid-init',

            # Python scripts
            'src/todiscgui',
            'src/tovid-stats',
            'src/titleset-wizard',
            'src/set_chapters',

            # Icons used in the GUIs
            'icons/hicolor/128x128/apps/tovid.png',
            'icons/hicolor/128x128/apps/titleset-wizard.png',

            # Config file
            'src/tovid.ini',
        ]),
        # Manual page
        ('share/man/man1',
         ['docs/man/tovid.1']),
        # Desktop shortcut
        ('share/applications',
         ['tovidgui.desktop',
         'titleset-wizard.desktop']),
        # Icons
        ('share/icons/hicolor/scalable/apps',
         [
             'icons/hicolor/scalable/apps/tovid.svg',
             'icons/hicolor/scalable/apps/titleset-wizard.svg',
             'icons/hicolor/scalable/apps/tovid_bw.svg',
             'icons/hicolor/scalable/apps/disc.svg',
             'icons/hicolor/scalable/apps/cd.svg',
         ]),
        ('share/icons/hicolor/128x128/apps',
         [
             'icons/hicolor/128x128/apps/tovid.png',
             'icons/hicolor/128x128/apps/titleset-wizard.png',
         ]),
        ('share/icons/hicolor/64x64/apps',
         [
             'icons/hicolor/64x64/apps/tovid.png',
             'icons/hicolor/64x64/apps/titleset-wizard.png',
         ]),
        ('share/icons/hicolor/48x48/apps',
         [
             'icons/hicolor/48x48/apps/tovid.png',
             'icons/hicolor/48x48/apps/titleset-wizard.png',
         ]),
        ('share/icons/hicolor/32x32/apps',
         [
             'icons/hicolor/32x32/apps/tovid.png',
             'icons/hicolor/32x32/apps/titleset-wizard.png',
         ]),
    ]
)




# idvid.py

from libtovid.metagui import *

_terse = Flag('Terse', '-terse')
_verbose = Flag('Verbose', '-verbose')
_fast = Flag('Fast', '-fast')
_tabular = Flag('Tabular', '-tabular')
_isformat = Choice('Is format', '-isformat', 'dvd',
    'Check the video for compliance with the given format',
    choices='vcd|svcd|dvd',
    toggles=True)
_istvsys = Choice('Is TV system', '-istvsys', 'ntsc',
    'Check whether the video is compliant with the given TV system',
    choices='ntsc|pal',
    toggles=True)

VIDEO_FILES = List('Video files', '', None,
    'List of multimedia video files to identify',
    Filename())

MAIN = VPanel('Main',
    _terse,
    _verbose,
    _fast,
    _tabular,
    _isformat,
    _istvsys,
    VIDEO_FILES,
)

def run():
    app = Application('idvid', MAIN)
    gui = GUI('idvid gui', 400, 600, app)
    gui.run()

if __name__ == '__main__':
    run()




"""This module provides an interface for running command-line applications.
Two primary classes are provided:

    `Command`
        For constructing and executing command-line commands
    `Pipe`
        For piping commands together

Commands are constructed by specifying a program to run, and each separate
argument to pass to that program. Arguments are used in a platform-independent
way, so there is no need to escape shell-specific characters or do any quoting
of arguments.

Commands may be executed in the foreground or background; they can print their
output on standard output, or capture it in a string variable.

For example::

    >>> echo = Command('echo', "Hello world")
    >>> echo.run()                                # doctest: +SKIP
    Hello world

Commands may be connected together with pipes::

    >>> sed = Command('sed', 's/world/nurse/')
    >>> pipe = Pipe(echo, sed)
    >>> pipe.run()                                # doctest: +SKIP
    Hello nurse

The above can be easily accomplished using Python standard library functions
like ``os.Popen`` or ``commands.getoutput``. This module was designed to
simplify the task of executing long-running commands or command-line pipes in
the background, and getting their output later. This behavior is controlled by
the arguments to the `~Command.run` method.

Normally, `~Command.run` prints all output on standard output. If instead you
need to capture the output, pass ``capture=True``, then retrieve it later with
`~Command.get_output`::

    >>> echo.run(capture=True)                    # doctest: +SKIP
    >>> echo.get_output()
    'Hello world\\n'

If the command you're running will take a long time to run, and you don't want
your program to be blocked while it's running, pass ``background=True``. If the
command may produce a lot of output during execution, you'll probably want to
capture it instead of printing it::

    >>> find = Command('find', '/')               # doctest: +SKIP
    >>> find.run(capture=True, background=True)

In this way, your application can keep doing other things while the long-running
command is executing; you can check whether it's done like so::

    >>> find.done()
    False
    >>> find.done()
    False
    >>> find.done()
    True

Then, as before, use `~Command.get_output` to get the output, if you need it. If
you need to get the standard error output, use `~Command.get_errors`.

"""
# Note: Some of the run() tests above will fail doctest.testmod(), since output
# from Command subprocesses is not seen as real output by doctest. The current
# workaround is to use the "doctest: +SKIP" directive (new in python 2.5).
# For other directives see http://www.python.org/doc/lib/doctest-options.html

# TODO: Make the 'capture/get_output' mechanisms able to handle large amounts
# of output, perhaps by logging to a temporary file and/or returning an iterator
# for lines of output, instead of a flat string.

__all__ = [
    'Command',
    'Pipe',
]

import subprocess
import signal
import os
# Small workaround for Python 3.x
from libtovid import unicode, basestring


class ProgramNotFound (ValueError):
    """Raised when the program given to a command is not available.
    """
    pass

class Command:
    """A command-line statement, consisting of a program and its arguments,
    with support for various modes of execution.
    """
    def __init__(self, program, *args):
        """Create a Command to run a program with the given arguments.

            program
                A string containing the name of a program to execute
            args
                Individual arguments to supply the command with

        For example::

            >>> cmd = Command('echo', 'Hello world')

        """
        self.program = program
        self.args = []
        for arg in args:
            self.add(arg)
        self.proc = None
        self.output = ''
        self.error = ''


    def add(self, *args):
        """Append one or more arguments to the command. Each argument passed
        to this function is converted to string form, and treated as a single
        argument in the command. No special quoting or escaping of argument
        contents is necessary.
        """
        for arg in args:
            self.args.append(unicode(arg))


    def run(self, capture=False, background=False, silent=False):
        """Run the command and capture or display output.

            capture
                ``False`` to show command output/errors on stdout,
                ``True`` to capture output/errors for retrieval
                by `get_output` and `get_errors`
            background
                ``False`` to wait for command to finish running,
                ``True`` to run process in the background
            silent
                ``False`` to print each command as it runs,
                ``True`` to run silently

        By default, this function displays all command output, and waits
        for the program to finish running, which is usually what you'd want.
        Capture output if you don't want it printed immediately (and call
        `get_output` later to retrieve it).

        This function does not allow special stream redirection. For that,
        use `run_redir`.
        """
        if not silent:
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            print("Running command:")
            print(unicode(self))
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

        if capture:
            self.run_redir(None, subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            self.run_redir(None, None, stderr=None)
        if not background:
            self.wait()


    def run_redir(self, stdin=None, stdout=None, stderr=None):
        """Execute the command using the given stream redirections.

            stdin
                Filename or `file` object to read input from
            stdout
                Filename or `file` object to write output to
            stderr
                Filename or `file` object to write errors to

        Use ``None`` for regular system stdin/stdout/stderr (default behavior).
        That is, if ``stdout=None``, the command's standard output is printed.

        This function is used internally by `run`; if you need to do
        stream redirection (ex. ``spumux < menu.mpg > menu_subs.mpg``), use
        this function instead of `run`, and call `wait` afterwards
        if needed.
        """
        self.output = ''
        # Open files if string filenames were provided
        if isinstance(stdin, basestring):
            stdin = open(stdin, 'r')
        if isinstance(stdout, basestring):
            stdout = open(stdout, 'w')
        if isinstance(stderr, basestring):
            stderr = open(stderr, 'w')
        # Run the subprocess
        try:
            self.proc = subprocess.Popen([self.program] + self.args,
                              stdin=stdin, stdout=stdout, stderr=stderr)
        except OSError:
            raise ProgramNotFound("Program '%s' not found." % self.program)


    def wait(self):
        """Wait for the command to finish running, and return the result
        (`self.proc.returncode`).

        If a :exc:`KeyboardInterrupt` occurs (user pressed Ctrl-C), the
        subprocess is killed (and :exc:`KeyboardInterrupt` re-raised).
        """
        if not isinstance(self.proc, subprocess.Popen):
            print("**** Can't wait(): Command is not running")
            return
        try:
            result = self.proc.wait()
        except KeyboardInterrupt:
            self.kill()
            raise KeyboardInterrupt
        return result


    def kill(self):
        """Abort!
        """
        os.kill(self.proc.pid, signal.SIGTERM)
        #self.proc.kill()



    def done(self):
        """Return ``True`` if the command is finished running; ``False``
        otherwise. Only useful if the command is run in the background.
        """
        return self.proc.poll() != None


    def get_output(self):
        """Wait for the command to finish running, and return a string
        containing the command's output. If this command is piped into another,
        return that command's output instead. Returns an empty string if the
        command has not been run yet.
        """
        if self.output == '' and isinstance(self.proc, subprocess.Popen):
            self.output = self.proc.communicate()[0]
        return self.output


    def get_errors(self):
        """Wait for the command to finish running, and return a string
        containing the command's standard error output. Returns an empty
        string if the command has not been run yet.
        """
        if self.error == '' and isinstance(self.proc, subprocess.Popen):
            self.error = self.proc.communicate()[1]
        return self.error


    def __str__(self):
        """Return a string representation of the Command, as it would look if
        run in a command-line shell.
        """
        ret = self.program
        for arg in self.args:
            ret += " %s" % _enc_arg(arg)
        return ret


    def to_script(self):
        """Return a bash script for running the given command.
        """
        script = '#!/usr/bin/env bash' + '\n\n'
        script_path = os.environ['PATH'].rsplit(':')
        script += 'PATH=' + script_path[0] + ':$PATH'  + '\n\n'
        # Write arguments, one per line with backslash-continuation
        words = [_enc_arg(arg) for arg in [self.program] + self.args]
        script += ' \\\n'.join(words)
        return script


class Pipe:
    """Several `Command` objects, each having its output piped into the next.
    """
    def __init__(self, *commands):
        """Create a new Pipe containing all the given Commands."""
        self.commands = []
        for cmd in commands:
            self.add(cmd)
        self.proc = None
        self.output = ''


    def add(self, *commands):
        """Append the given commands to the end of the pipeline."""
        for cmd in commands:
            self.commands.append(cmd)


    def run(self, capture=False, background=False):
        """Run all Commands in the pipeline, doing appropriate stream
        redirection for piping.

            capture
                ``False`` to show pipeline output on stdout,
                ``True`` to capture output for retrieval by `get_output`

        """
        self.output = ''
        prev_stdout = None
        # Run each command, piping to the next
        for cmd in self.commands:
            # If this is not the last command, pipe into the next one
            if cmd != self.commands[-1]:
                cmd.run_redir(prev_stdout, subprocess.PIPE, None)
                prev_stdout = cmd.proc.stdout
            # Last command in pipeline; direct output appropriately
            else:
                if capture:
                    cmd.run_redir(prev_stdout, subprocess.PIPE, None)
                else:
                    cmd.run_redir(prev_stdout, None, None)
                # Wait for last command to finish
                if not background:
                    cmd.wait()


    def get_output(self):
        """Wait for the pipeline to finish executing, and return a string
        containing the output from the last command in the pipeline.
        """
        return self.commands[-1].get_output()


    def __str__(self):
        """Return a string representation of the Pipe.
        """
        commands = [unicode(cmd) for cmd in self.commands]
        return ' | '.join(commands)


def _enc_arg(arg):
    """Quote an argument for proper handling of special shell characters.
    Don't quote unless necessary. For example:

        >>> print(_enc_arg("spam"))
        spam
        >>> print(_enc_arg("spam & eggs"))
        'spam & eggs'
        >>> print(_enc_arg("['&&']"))
        '['\\''&&'\\'']'

    This is used internally by Command; you'd only need this if you're running
    shell programs without using the Command class.
    """
    arg = unicode(arg)
    # At the first sign of any special character in the argument,
    # single-quote the whole thing and return it (escaping ' itself)
    for char in ' #"\'\\&|<>()[]!?*':
        if char in arg:
            return "'%s'" % arg.replace("'", "'\\''")
    # No special characters found; use literal string
    return arg






"""GUI widgets for grouping Controls together and defining layout.
"""

__all__ = [
    'Label',
    'Panel',
    'HPanel',
    'VPanel',
    'Dropdowns',
    'Drawer',
    'Tabs',
    'FlagGroup',
]

# Python < 3.x
try:
    import Tkinter as tk
# Python 3.x
except ImportError:
    import tkinter as tk

from base64 import b64encode
from libtovid import cli
from libtovid.odict import Odict
from libtovid.metagui.widget import Widget
from libtovid.metagui.control import Control, Flag
from libtovid.metagui.variable import ListVar
from libtovid.metagui.support import (
  ComboBox, ensure_type, divide_list, get_photo_image
  )


class Label (Widget):
    """A widget with a text label.
    """
    def __init__(self,
                 text='',
                 justify='left',
                 image_file='',
                 image_width=0,
                 image_height=0):
        """Create a Label with the given text.

            text
                String to appear in the Label
            justify
                Text justification: 'left', 'center', or 'right'
            image_file
                Full path to image file. May be in any format that
                'convert' supports.
            image_width
                Width in pixels to resize the image to
            image_height
                Height in pixels to resize the image to

        Width and height may be used to scale the image in different ways:

            width == 0, height == 0
                Preserve the original image's size
            width > 0, height == 0
                Resize to the given width; height automatically
                adjusts to maintain aspect ratio
            width == 0, height > 0
                Resize to the given height; width automatically
                adjusts to maintain aspect ratio
            width > 0, height > 0
                Resize to exactly the given dimensions
        """
        Widget.__init__(self, text)
        self.text = text
        if justify not in ['left', 'center', 'right']:
            raise ValueError("Label justify argument must be 'left', 'center', "
                             "or 'right' (got '%s' instead)." % justify)
        self.justify = justify
        # Image attributes
        self.image_file = image_file
        self.image_width = image_width
        self.image_height = image_height
        # Will be set by draw()
        self.label = None
        self.image = None


    def draw(self, master, **kwargs):
        """Draw the Label in the given master.
        """
        Widget.draw(self, master, **kwargs)
        # If an image filename was provided, get a PhotoImage
        if self.image_file:
            photo_image = get_photo_image(self.image_file,
                self.image_width, self.image_height, self.cget('background'))
            # Keep a reference to the PhotoImage to prevent garbage collection
            self.photo = photo_image
            image_frame = tk.Frame(self, padx=4, pady=4)
            self.image = tk.Label(image_frame, image=photo_image)
            self.image.pack()
            image_frame.pack(side='left', expand=False)

        # Create and pack the label
        self.label = tk.Label(self, text=self.text, justify=self.justify, padx=8)
        # Set appropriate anchoring based on justification
        _anchors = {'left': 'w', 'center': 'center', 'right': 'e'}
        self.label.pack(expand=True, anchor=_anchors[self.justify])


class Panel (Widget):
    """A group of Widgets in a rectangular frame, with an optional label.
    """
    def __init__(self, name='', *widgets, **kwargs):
        """Create a Panel containing one or more widgets or sub-panels.

            name
                Name (label) for panel, or '' for no label
            widgets
                One or more Widgets (Controls, Panels, Drawers etc.)
        """
        Widget.__init__(self, name, **kwargs)

        # Ensure that Widgets were provided
        ensure_type("Panels may only contain Widgets.", Widget, *widgets)
        self.widgets = list(widgets)
        # Defined by draw()
        self.frame = None


    def draw(self, master, **kwargs):
        """Draw the Panel, but not any contained widgets.

            labeled
                True (default) to use a LabelFrame with the panel's name,
                False to draw the panel without a label.

        Panel subclasses must pack the contained widgets, or
        call ``draw_widgets`` to pack them.
        """
        Widget.draw(self, master, **kwargs)
        # Get a labeled or unlabeled frame (labeled by default)
        if self.name and kwargs.get('labeled', True):
            self.frame = tk.LabelFrame(self, text=self.name,
                                       padx=4, pady=4)
        else:
            self.frame = tk.Frame(self)
        # Pack the frame
        self.frame.pack(side='top', fill='both', expand=True)


    def draw_widgets(self, side='top', expand=False):
        """Draw contained widgets in self.frame.

            side
                'top' or 'left', to pack widgets on that side
            expand
                True to make widgets fill all available space

        Panel subclasses may call this method to pack widgets,
        or do custom packing (and not call this method).
        """
        for widget in self.widgets:
            widget.draw(self.frame)
            widget.pack(side=side, expand=expand, anchor='nw',
                        fill='both', padx=4, pady=2)


    def get_args(self):
        """Return a list of all command-line options from contained widgets.
        """
        args = []
        for widget in self.widgets:
            args += widget.get_args()
        return args


    def set_args(self, args):
        """Set panel options from the given list of command-line arguments,
        and remove any successfully parsed options and arguments from ``args``.
        """
        for widget in self.widgets:
            widget.set_args(args)


    def enable(self, enabled=True):
        """Enable all widgets in the Panel.
        """
        for widget in self.widgets:
            widget.enable(enabled)


class HPanel (Panel):
    """A group of widgets or sub-panels, packed horizontally (left-to-right).

    For example::

        HPanel("General",
            Filename(...),
            Flag(...),
            Number(...)
            )
    """
    def __init__(self, name='', *widgets, **kwargs):
        """Create an HPanel to show the given widgets in a horizontal layout.
        """
        Panel.__init__(self, name, *widgets, **kwargs)


    def draw(self, master, **kwargs):
        """Draw the HPanel and its contained widgets in the given master.
        """
        Panel.draw(self, master, **kwargs)
        self.draw_widgets(side='left', expand=True)


class VPanel (Panel):
    """A group of widgets or sub-panels, packed vertically (top-to-bottom).

    For example::

        VPanel("General",
            Filename(...),
            Flag(...),
            Number(...)
            )
    """
    def __init__(self, name='', *widgets, **kwargs):
        """Create a VPanel to show the given widgets in a vertical layout.
        """
        Panel.__init__(self, name, *widgets, **kwargs)


    def draw(self, master, **kwargs):
        """Draw the VPanel and its contained widgets in the given master.
        """
        Panel.draw(self, master, **kwargs)
        # Avoid unnecessary vertical expansion
        self.draw_widgets(side='top', expand=False)


class Dropdowns (Panel):
    """A Panel with Controls that are shown/hidden using a dropdown list.

    The Dropdowns panel initially displays a single dropdown list, with
    one entry for each Control. When a dropdown entry is selected, the
    corresponding Control is displayed, along with a "remove" button to
    discard the control.
    """
    def __init__(self, name='', *controls, **kwargs):
        Panel.__init__(self, name, *controls, **kwargs)
        ensure_type("Dropdown contents must be Controls", Control, *controls)
        # Controls, indexed by label
        self.controls = Odict()
        for control in self.widgets:
            self.controls[control.label] = control
        # Set by draw()
        self.choices = None
        self.chosen = None
        self.chooser = None


    def draw(self, master, **kwargs):
        """Draw the Dropdowns widget in the given master.
        """
        Panel.draw(self, master, **kwargs)
        # List of Controls and chosen item
        self.choices = ListVar(items=self.controls.keys())
        self.chosen = tk.StringVar()
        # ComboBox to choose Controls from
        self.chooser = ComboBox(self.frame, self.choices, variable=self.chosen,
                                command=self.choose_new)
        self.chooser.pack(anchor='nw')


    def choose_new(self, event=None):
        """Create and display the chosen control."""
        chosen = self.chosen.get()
        if chosen == '':
            return
        self.chooser.pack_forget()
        # Put control and remove button in a frame inside self.frame
        control_frame = tk.Frame(self.frame)
        button = tk.Button(control_frame, text="X",
                           command=lambda:self.remove(chosen))
        button.pack(side='left')
        control = self.controls[chosen]
        control.draw(control_frame)
        control.pack(side='left', fill='x', expand=True)
        control_frame.pack(fill='x', expand=True)
        # Remove the chosen control/panel from the list of available ones
        self.choices.remove(chosen)
        self.chooser.pack(anchor='nw')


    def remove(self, control_label):
        """Remove a given Control from the interface."""
        frame = self.controls[control_label].master
        frame.pack_forget()
        frame.destroy()
        # Make this control available in the dropdown again
        self.choices.append(control_label)


    def get_args(self):
        """Return a list of command-line options from all active Controls.
        """
        args = []
        for control in self.widgets:
            if control.is_drawn:
                args += control.get_args()
        return args


class Drawer (Panel):
    """A Panel that may be hidden or "closed" like a drawer.
    """
    def __init__(self, name='', *widgets, **kwargs):
        """Create a Drawer containing the given widgets.
        """
        Panel.__init__(self, name, *widgets, **kwargs)
        self.visible = False
        # Set by draw()
        self.button = None


    def draw(self, master, **kwargs):
        """Draw the Drawer, with contained widgets initially hidden.
        """
        # Draw the base panel and contained widgets
        Panel.draw(self, master, **kwargs)
        self.frame.pack_forget()
        self.draw_widgets()
        # Add a checkbutton for showing/hiding
        self.button = tk.Button(self, text=self.name, relief='groove',
                           command=self.show_hide)
        self.button.pack(anchor='nw', fill='x', expand=True)


    def show_hide(self):
        """Show or hide the widgets in the Drawer.
        """
        # Hide if showing
        if self.visible:
            self.frame.pack_forget()
            self.visible = False
            self.button.config(relief='groove')
        # Show if hidden
        else:
            self.frame.pack(anchor='nw', fill='both', expand=True)
            self.visible = True
            self.button.config(relief='sunken')


class Tabs (Panel):
    """A Panel with tab buttons that switch between several widgets.
    """
    def __init__(self, name='', *widgets, **kwargs):
        """Create a tabbed panel that switch between several widgets.
        """
        Panel.__init__(self, name, *widgets, **kwargs)
        # Index of selected tab
        self.index = 0
        self.side = kwargs.get('side', 'top')
        # Set by draw()
        self.buttons = None
        self.selected = None


    def draw(self, master, **kwargs):
        """Draw the Tabs widget in the given master.
        """
        Panel.draw(self, master, **kwargs)
        # Selected tab index
        self.selected = tk.IntVar()
        # Tkinter configuration common to all tab buttons
        config = {
            'variable': self.selected,
            'command': self.change,
            'selectcolor': 'white',
            'relief': 'sunken',
            'offrelief': 'groove',
            'indicatoron': 0,
            'padx': 4, 'pady': 4
            }
        # Frame to hold tab buttons
        self.buttons = tk.Frame(self.frame)
        # For tabs on left or right, pack tab buttons vertically
        if self.side in ['left', 'right']:
            button_side = 'top'
            bar_anchor = 'n'
            bar_fill = 'y'
        else:
            button_side = 'left'
            bar_anchor = 'w'
            bar_fill = 'x'
        # Tab buttons, numbered from 0
        for index, widget in enumerate(self.widgets):
            button = tk.Radiobutton(self.buttons, text=widget.name,
                                    value=index, **config)
            button.pack(anchor='nw', side=button_side,
                        fill='both', expand=True)
            # For Panels, hide the panel's own label
            if isinstance(widget, Panel):
                widget.draw(self.frame, labeled=False)
            else:
                widget.draw(self.frame)
        self.buttons.pack(anchor=bar_anchor, side=self.side, fill=bar_fill)
        # Activate the first tab
        self.selected.set(0)
        self.change()


    def change(self):
        """Event handler for switching tabs
        """
        # Unpack the existing widget
        self.widgets[self.index].pack_forget()
        # Pack the newly-selected widget
        selected = self.selected.get()
        self.widgets[selected].pack(side=self.side, fill='both', expand=True)
        # Remember this tab's index
        self.index = selected


    def activate(self, tab):
        """Display the given tab (either by index or Widget instance).
        """
        # Get index by Widget instance
        if isinstance(tab, Widget):
            if tab in self.widgets:
                index = list(self.widgets).index(tab)
            else:
                raise ValueError("Widget '%s' not in Tabs" % tab.name)
        # Or use index directly
        else:
            index = int(tab)
        # Switch to the given tab index
        self.selected.set(index)
        self.change()


class FlagGroup (Panel):
    """A Panel that shows a group of several Flag controls,
    optionally making them mutually-exclusive.
    """
    def __init__(self,
                 name='',
                 kind='normal',
                 *flags,
                 **kwargs):
        """Create a FlagGroup with the given label and state.

            name
                Name/label for the group
            kind
                'normal' for independent Flags, 'exclusive' for
                mutually-exclusive Flags (more like a Choice)
            flags
                One or more Flag controls to include in the group

        These keyword arguments are accepted:

            side
                'top' or 'left', to pack Flags vertically or horizontally
            rows
                For 'left' packing, number of rows to split flags into
            columns
                For 'top' packing, number of columns to split flags into
        """
        Panel.__init__(self, name, **kwargs)
        ensure_type("FlagGroup may only contain Flag instances", Flag, *flags)
        self.flags = flags
        self.kind = kind
        # Keyword arguments
        self.side = kwargs.get('side', 'top')
        if self.side not in ['top', 'left']:
            raise ValueError("FlagGroup 'side' must be 'top' or 'left'.")
        self.columns = kwargs.get('columns', 1)
        self.rows = kwargs.get('rows', 1)


    def draw(self, master, **kwargs):
        """Draw the FlagGroup in the given master widget.
        """
        Panel.draw(self, master, **kwargs)

        if self.side == 'top':
            num_strips = self.columns
            strip_side = 'left'
        else:
            num_strips = self.rows
            strip_side = 'top'

        # Divide flags into rows or columns
        for flags in divide_list(self.flags, num_strips):
            subframe = tk.Frame(self.frame)
            # Draw all Flags in the row/column
            for flag in flags:
                flag.draw(subframe)
                flag.add_callback(self.modified)
                flag.pack(anchor='nw', side=self.side, fill='x', expand=True)
            # Pack the frame for the current row/column
            subframe.pack(anchor='nw', side=strip_side)


    def modified(self, flag):
        """Called when a flag is modified.
        """
        # For normal flags, do nothing
        if self.kind != 'exclusive':
            return
        # If the flag is unchecked, do nothing
        if not flag.get():
            return
        # Otherwise, uncheck all other flags
        for _flag in self.flags:
            if _flag != flag:
                _flag.set(False)


    def get_args(self):
        """Return a list of arguments for setting the relevant flag(s).
        """
        args = []
        for flag in self.flags:
            # Include only if set to a non-default value
            if flag.option != 'none' and flag.get() != flag.default:
                args.extend(flag.get_args())
        return args






__all__ = [
    # Functions
    'escape',
    'float_to_ratio',
    'get_code_lines',
    'get_file_type',
    'indent_level',
    'pretty_dict',
    'ratio_to_float',
    'safe_filename',
    'tokenize',
    'temp_dir',
    'temp_file',
    'to_unicode',
    'trim',
    'wait',
    'imagemagick_version',
    'imagemagick_fonts',
]

import os
import sys
import shlex
import doctest
import mimetypes
import tempfile
# Python 3 compatibility
from libtovid import basestring, unicode

# Special characters that may need escaping
special_chars = '\\ #*:;&?!<>[]()"\''

def safe_filename(filename, work_dir):
    """Ensure the given ``filename`` is free of quirky or problematic characters.
    If so, simply return ``filename``; if not, create a safe symlink in
    ``work_dir`` to the original file, and return the symlink name.
    """
    safename = filename
    for char in special_chars:
        safename = safename.replace(char, '_')
    if safename != filename:
        basename = os.path.basename(safename)
        safename = os.path.join(work_dir, basename)
        os.symlink(filename, safename)
    return safename


def escape(text):
    """Return a copy of the given text string with potentially problematic
    "special" characters backslash-escaped.
    """
    result = text
    for char in special_chars:
        result = result.replace(char, '\%s' % char)
    return result


def to_unicode(text):
    """Return the given text string, converted to unicode.
    """
    if not isinstance(text, unicode):
        text = unicode(str(text).decode('latin-1'))
    return text


def indent_level(line):
    """Return the number of leading whitespace characters in the line.
    Tab, newline, and carriage-return characters each count as a single space.

    For example::

        >>> indent_level('      foo')
        6
        >>> indent_level('foo')
        0
        >>> indent_level('\\r\\n\\t foo')
        4

    """
    return len(line) - len(line.lstrip())


def trim(text):
    """Strip leading indentation from a block of text.
    Borrowed from `PEP 257`_.
    
    .. _PEP 257: http://www.python.org/peps/pep-0257.html
    """
    if not text:
        return ''
    # Split text into lines, converting tabs to spaces
    lines = text.expandtabs().splitlines()
    # Determine minimum indentation (except first line)
    try:
        MAXINT = sys.maxint
    # python 3 compatibility
    except AttributeError:
        MAXINT = sys.maxsize

    indent = MAXINT

    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special)
    trimmed = [lines[0].strip()]
    if indent < MAXINT:
        for line in lines[1:]:
            # Append line, minus indentation
            trimmed.append(line[indent:].rstrip())
    # Strip leading blank lines
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Strip trailing blank lines
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    # Return a string, rejoined with newlines
    return '\n'.join(trimmed)


def ratio_to_float(ratio):
    """Convert a string expressing a numeric ratio, with X and Y parts
    separated by a colon ':', into a floating-point decimal number.

    For example::

        >>> ratio_to_float('4:3')
        1.3333333333333333

        >>> ratio_to_float('16:9')
        1.7777777777777777

    """
    values = ratio.split(':', 1)
    if len(values) == 2:
        return float(values[0]) / float(values[1])
    elif len(values) == 1:
        return float(values[0])
    else:
        raise ValueError("ratio_to_float: too many values in ratio '%s'" % ratio)


def float_to_ratio(number):
    """Convert a decimal number into an integer ratio string 'X:Y'.
    Keeps three digits of precision. No attempt is made to find a
    greatest common divisor.

    For example::

        >>> float_to_ratio(1.3333333333333333)
        '1333:1000'

        >>> float_to_ratio(16.0 / 9)
        '1777:1000'

    """
    numerator = int(number * 1000)
    return "%d:1000" % numerator


def tokenize(line, include_chars=''):
    """Separate a text line into tokens, returning them in a list. By default,
    tokens are space-separated, and each token consists of ``[a-z]``,
    ``[A-Z]``, ``[0-9]``, or any of ``.:-%()/``. Additional valid token
    characters may be specified by passing them in the include_chars string.
    """
    lexer = shlex.shlex(line, posix = True)
    # Rules for splitting tokens
    lexer.wordchars = lexer.wordchars + '.:-%()/' + include_chars
    lexer.whitespace_split = False
    # Append all tokens to a list
    tokens = []
    while True:
        token = lexer.get_token()
        if not token:
            break
        else:
            tokens.append(token)
    return tokens


def pretty_dict(a_dict):
    """Return a pretty-printed dictionary, with one line for each key.
    Keys are printed in sorted ascending order.
    """
    result = ''
    keys = a_dict.keys()
    keys.sort()
    for key in keys:
        value = a_dict[key]
        # For boolean options, print Trues and omit Falses
        if value.__class__ == bool:
            if value is True:
                result += "    %s\n" % key
            else:
                pass
        # If value has spaces, quote it
        elif value.__class__ == str and ' ' in value:
            result += "    %s \"%s\"\n" % (key, value)
        # Otherwise, don't
        else:
            result += "    %s %s\n" % (key, value)
    return result


def get_code_lines(filename):
    """Return a list of all lines of code in the given file.
    Whitespace and lines beginning with ``#`` are ignored.
    """
    infile = open(filename, 'r')
    codelines = []
    for line in infile.readlines():
        if line.lstrip() and not line.lstrip().startswith('#'):
            codelines.append(line)
    infile.close()
    return codelines


def get_file_type(filename):
    """Return 'image', 'audio', or 'video', if the given filename appears to be
    any of those types; otherwise, return ``None``. Determined by file's mimetype,
    which is based on filename extension, so possibly inaccurate. Returns
    ``None`` for any directory or extensionless filename.
    """
    mimetype, encoding = mimetypes.guess_type(filename)
    # Get the base type (the part before '/')
    basetype = None
    if mimetype:
        basetype = mimetype[0:mimetype.find('/')]
    if basetype in ['image', 'audio', 'video']:
        return basetype
    else:
        return None


def temp_dir(prefix=''):
    """Create a unique temporary directory and return its full pathname.
    """
    import libtovid
    work_dir = libtovid.Config().get('DEFAULT', 'work_dir')
    return tempfile.mkdtemp(prefix="%s_" % prefix,
                            dir=work_dir)


def temp_file(prefix='', suffix=''):
    """Create a unique temporary file and return its full pathname.
    """
    import libtovid
    work_dir = libtovid.Config().get('DEFAULT', 'work_dir')
    fd, fname = tempfile.mkstemp(prefix="%s_" % prefix,
                                 suffix=suffix,
                                 dir=work_dir)
    os.close(fd)
    return fname


def wait(seconds):
    """Print a message and pause for the given number of seconds.
    """
    print("Resuming in %s seconds..." % seconds)
    os.system('sleep %s' % seconds)


def imagemagick_version():
    """Return the version of ImageMagick that's currently installed,
    as a list of integers (ex. ``[6, 3, 5, 10]`` for version 6.3.5.10).
    """
    command = 'convert -list configure | grep ^LIB_VERSION_NUMBER'
    lines = os.popen(command).readlines()
    version_string = lines[0].split()[1]
    return [int(n) for n in version_string.split(',')]


def imagemagick_fonts():
    """Return a list of fonts available to ImageMagick.
    """
    version = imagemagick_version()

    # ImageMagick 6.4.x prints fonts like this:
    #   Font: Verdana-Italic
    #     family: Verdana
    #     style: Italic
    #     stretch: Normal
    #     weight: 400
    if version >= [6, 4, 0, 0]:
        command = "convert -list font | grep 'Font:' | awk '{print $2}'"

    # ImageMagick 6.3.5.7 and later print fonts like this:
    #   Name             Family       Style  Stretch Weight
    #   ---------------------------------------------------
    #   Verdana-Italic   Verdana      Italic  Normal    400
    elif version >= [6, 3, 5, 7]:
        command = "convert -list font | awk '!/\-\-\-|Path|Name|^$/ {print $1}'"

    # Older versions use 'type' instead of 'font', same formatting as above
    else:
        command = "convert -list type | awk '!/\-\-\-|Path|Name|^$/ {print $1}'"

    # Run the command and return the list of font names
    lines = os.popen(command).readlines()
    return sorted([line.rstrip('\n') for line in lines \
        if not 'undefin' in line.lower()])





"""This module is a GUI builder with a GUI. It generates metagui code for
a specific program, with user interaction for fine-tuning the metagui widgets.

Ideas:

Parse manpage to get usage/option information, ex.:

    -audio-demuxer <[+]name> (-audiofile only)
    -audiofile <filename>
    -audiofile-cache <kBytes>
    -bandwidth <value> (network only)

User-customizable regex will ignore (..), create Controls for <..> etc.

Or: Keep a collection of "samples" of every time an option
is mentioned (anywhere in the manpage), and base widget config
decisions on the samples.

Need a GUI for making the process interactive. i.e. "I found these options
in the manpage, but you can make adjustments if necessary":

    [ ] -option1 [Number[^]] [Label] [default]
        [Tooltip text________________________]
        [min] [max] [scale[^]]

    [x] -option2 [Choice[^]] [Label] [default]
        [Tooltip text________________________]
        [a|b|c]

[Number[^]] is a combobox for choosing control type; [min] [max] [scale[^]]
and [a|b|c] are control-configuration widgets that appear depending on the
chosen control type. Checking an option includes it in the metaGUI.
"""

__all__ = [
    'ControlEditor',
    'ControlChooser',
]

from inspect import getargspec

# Python < 3.x
try:
    import Tkinter as tk
# Python 3.x
except ImportError:
    import tkinter as tk

from libtovid.odict import Odict
from libtovid.metagui import Text, Number, Flag
from libtovid.metagui.control import CONTROLS
from libtovid.metagui.support import ComboBox

class ControlEditor (tk.Frame):
    """A GUI control panel for setting Control arguments."""
    def __init__(self, control):
        """Create a GUI panel for the given Control subclass.
        """
        self.control = control
        # Find out what arguments/defaults the Control constructor has
        args, varargs, varkw, defaults = getargspec(self.control.__init__)
        self.kwargs = Odict(args[1:], defaults)

    def draw(self, master):
        """Draw widgets for adjusting all Control.__init__ arguments."""
        tk.Frame.__init__(self, master)
        for arg, default in self.kwargs.items():
            if type(default) == str:
                widget = Text('', arg, default)
            elif type(default) == int:
                widget = Number('', arg, default)
            elif type(default) == bool:
                widget = Flag('', arg, default)
            else:
                widget = Text('', arg, default)

            widget.draw(self)
            widget.pack()


class ControlChooser (tk.Frame):
    """A GUI panel for choosing a Control type and setting its arguments."""
    def __init__(self):
        self.control = tk.StringVar()

    def draw(self, master):
        tk.Frame.__init__(self, master)
        self.choices = ComboBox(self, CONTROLS.keys(),
                                variable=self.control,
                                command=self.refresh)
        self.choices.pack()
        self.editor = ControlEditor(Number)
        self.control.set('Number')
        self.editor.draw(self)
        self.editor.pack()

    def refresh(self, event=None):
        """Show the editor for the currently selected control.
        """
        self.editor.pack_forget()
        newcontrol = CONTROLS[self.control.get()]
        self.editor = ControlEditor(newcontrol)
        self.editor.draw(self)
        self.editor.pack()


# Demo
if __name__ == '__main__':
    root = tk.Tk()
    chooser = ControlChooser()
    chooser.draw(root)
    chooser.pack()
    root.mainloop()




"""Defines several commonly-used multimedia and image file types.
"""

__all__ = [
    'match_types',
    'get_extensions',
    'all_files',
    'image_files',
    'video_files',
    'audio_files',
]

import os
import mimetypes


def etc_mimetypes():
    """Get mimetypes from ``/etc/mime.types`` and return a dict of ``{ext: typename}``
    in the same format as returned by `mimetypes.types_map`.
    """
    if not os.path.exists('/etc/mime.types'):
        return {}
    mime_types = {}
    for line in open('/etc/mime.types', 'r'):
        if not line.startswith('#'):
            parts = line.split()
            if len(parts) > 1:
                for ext in parts[1:]:
                    mime_types['.' + ext] = parts[0]
    return mime_types


def match_types(containing):
    """Return a list of ``(type, extensions)`` tuples for matching mimetypes.

        containing
            String or list of strings to match

    For example::

        match_types('image/jpeg')

    Uses a "contains" search, so you can do::

        match_types(['image', 'mpeg'])

    to match any mimetype containing 'image' or 'mpeg'.

    The returned tuples are suitable for use as the 'filetypes' argument of
    Tkinter file dialogs like `askopenfilename`.
    """
    if type(containing) == str:
        containing = [containing]
    elif type(containing) != list:
        raise TypeError("match_types requires a string or list argument.")

    types = {}
    # Check for matching types and remember their extensions
    for ext, typename in sorted(mimetypes.types_map.items()):
        for substr in containing:
            if substr in typename:
                # Append or insert extension
                if typename in types:
                    types[typename] += ' *' + ext
                else:
                    types[typename] = '*' + ext

    # Convert to tuple-list form
    return types.items()


def get_extensions(containing):
    """Return a space-separated string of all extensions for matching types.
    Like `match_types`, but only return the extensions.
    """
    type_dict = dict(match_types(containing))
    ext_list = type_dict.values()
    return ' '.join(ext_list)


def new_filetype(name, extensions):
    """Create a filetype tuple from a name and a list of extensions.
    extensions may be a space-separated string, list, tuple, or other iterable.

        >>> new_filetype('My filetype', 'foo bar baz')
        ('My filetype', '*.foo *.FOO *.bar *.BAR *.baz *.BAZ')

    """
    if type(extensions) == str:
        extensions = extensions.split(' ')
    ext_patterns = ('*.%s *.%s' % (ext.lower(), ext.upper()) for ext in extensions)
    return (name, ' '.join(ext_patterns))


all_files = ('All files', '*.*')
image_files = ('Image files', get_extensions('image'))
video_files = ('Video files', get_extensions('video'))
audio_files = ('Audio files', get_extensions('audio'))


